from typing import Dict
from typing import List

from compas.geometry import Scale
from compas.geometry import Transformation

from compas_ifc.entities.objectdefinition import ObjectDefinition
from compas_ifc.helpers import public_attributes


class Product(ObjectDefinition):
    """
    Class representing an IFC product. A product may be a physical object or a virtual object, that has one or more geometric representations.

    Attributes
    ----------
    axis : List[:class:`compas.geometry.Line`]
        The axis representation of the product. Converted to list of COMPAS lines.
    box : :class:`compas.geometry.Box`
        The box representation of the product. Converted to COMPAS box.
    body : List[:class:`compas_occ.geometry.Shape`]
        The body representation of the product. Converted to list of OCC shapes (BRep).
    body_with_opening : List[:class:`compas_occ.geometry.Shape`]
        The body representation of the product including openings. Converted to list of OCC shapes (BRep).
    opening : List[:class:`compas_occ`]
        The opening representation of the product. Converted to list of OCC shapes (BRep).
    transformation : :class:`compas.geometry.Transformation`
        The transformation of the product, Calculated from the stack of transformations of the product's parent entities. Read-only.
    style : dict
        The style of the product. Converted to dictionary.
    frame : :class:`compas.geometry.Frame`
        The frame of the product. Converted to COMPAS frame.


    """

    def __init__(self, entity, model) -> None:
        super().__init__(entity, model)
        self._axis = None
        self._box = None
        self._body = None
        self._opening = None
        self._body_with_opening = None
        self._transformation = None
        self._style = None
        self._frame = None

    def classifications(self) -> List[Dict[str, str]]:
        """
        External sources of information associated with this product.
        The source of information can be:

        * a classification system;
        * a dictionary server;
        * any external catalogue that classifies the product further;
        * a service that combines the above.

        Returns
        -------
        List[dict[str, str]]

        References
        ----------
        :ifc:`classification`

        """
        classifications = []
        if self._entity.HasAssociations:
            for association in self._entity.HasAssociations:
                if association.is_a("IfcRelAssociatesClassification"):
                    classifications.append(
                        {
                            "name": association.Name,
                            "identification": association.RelatingClassification.Identification,
                            "source": association.RelatingClassification.ReferencedSource.Source,
                        }
                    )
        return classifications

    def materials(self) -> List[Dict]:
        """
        The materials associated with this product.

        Returns
        -------
        List[Dict]

        """
        materials = []

        for association in self._entity.HasAssociations:
            if not association.is_a("IfcRelAssociatesMaterial"):
                continue

            if association.RelatingMaterial.is_a("IfcMaterialLayerSet"):
                print(public_attributes(association.RelatingMaterial))

            elif association.RelatingMaterial.is_a("IfcMaterialLayerSetUsage"):
                print(public_attributes(association.RelatingMaterial))

            elif association.RelatingMaterial.is_a("IfcMaterialConstituentSet"):
                # name = association.RelatingMaterial.Name
                for constituent in association.RelatingMaterial.MaterialConstituents:
                    for pset in constituent.Material.HasProperties:
                        properties = {}
                        for prop in pset.Properties:
                            key = prop.Name
                            value = prop.NominalValue.get_info()["wrappedValue"]
                            properties[key] = value

            else:
                raise NotImplementedError

        return materials

    @property
    def axis(self):
        from compas_ifc.representation import entity_axis_geometry

        if not self._axis:
            self._axis = entity_axis_geometry(self)
        return self._axis

    @property
    def box(self):
        from compas_ifc.representation import entity_box_geometry

        if not self._box:
            self._box = entity_box_geometry(self) or entity_box_geometry(self, context="Plan")
        return self._box

    @property
    def body(self):
        from compas_ifc.representation import entity_body_geometry

        if not self._body:
            self._body = entity_body_geometry(self, use_occ=self.model.reader.use_occ)
        return self._body

    @body.setter
    def body(self, value):
        self._body = value

    @property
    def opening(self):
        from compas_ifc.representation import entity_opening_geometry

        if not self._opening:
            self._opening = entity_opening_geometry(self, use_occ=self.model.reader.use_occ)
        return self._opening

    @opening.setter
    def opening(self, value):
        self._opening = value

    @property
    def body_with_opening(self):
        if self._body_with_opening is None:
            if self._entity:
                if not self._body_with_opening:
                    cached_geometry = self.model.reader.get_preloaded_geometry(self)
                    if cached_geometry:
                        self._body_with_opening = cached_geometry
                    else:
                        return
                        # TODO: double check if this is still triggered with preloaded geometry
                        # self._body_with_opening = entity_body_with_opening_geometry(self, use_occ=self.model.reader.use_occ)

            else:
                scale = self.model.project.length_scale
                T = Transformation.from_frame(self.frame)
                S = Scale.from_factors([scale, scale, scale])

                from compas.geometry import Shape

                if isinstance(self.body, Shape):
                    geometry = self.body.transformed(S * T)
                    geometry.scale(scale)
                else:
                    geometry = self.body.transformed(S * T)

                self._body_with_opening = geometry

        return self._body_with_opening

    @body_with_opening.setter
    def body_with_opening(self, value):
        self._body_with_opening = value

    @property
    def geometry(self):
        return self.body_with_opening

    @property
    def transformation(self):
        from compas_ifc.representation import entity_transformation

        if not self._transformation:
            self._transformation = entity_transformation(self)
        return self._transformation

    @property
    def frame(self):
        from compas_ifc.representation import entity_frame

        if not self._frame and self._entity:
            self._frame = entity_frame(self)
        return self._frame

    @frame.setter
    def frame(self, value):
        self._frame = value

    @property
    def style(self):
        if not self._style:
            if self._entity:
                self._style = self.model.reader.get_preloaded_style(self)
            else:
                self._style = {}
        # TODO: handle non-preloaded situation
        return self._style

    def show(self):
        self.model.show(self)



from functools import reduce
from operator import mul

from compas.geometry import Frame
from compas.geometry import Point
from compas.geometry import Transformation
from compas.geometry import Vector


def IfcLocalPlacement_to_transformation(placement, scale=1) -> Transformation:
    """
    Convert an IFC LocalPlacement [localplacement]_ to a COMPAS transformation.
    This will resolve all relative placements into one transformation wrt the global coordinate system.

    """
    stack = []
    while True:
        Location = placement.RelativePlacement.Location
        Axis = placement.RelativePlacement.Axis
        RefDirection = placement.RelativePlacement.RefDirection

        if Axis and RefDirection:
            zaxis = Vector(*Axis.DirectionRatios)
            xaxis = Vector(*RefDirection.DirectionRatios)
            yaxis = zaxis.cross(xaxis)
            xaxis = yaxis.cross(zaxis)
        else:
            xaxis = Vector.Xaxis()
            yaxis = Vector.Yaxis()

        point = Point(*Location.Coordinates) * scale
        frame = Frame(point, xaxis, yaxis)
        stack.append(frame)

        if not placement.PlacementRelTo:
            break

        placement = placement.PlacementRelTo

    matrices = [Transformation.from_frame(f) for f in stack]
    return reduce(mul, matrices[::-1])


def IfcLocalPlacement_to_frame(placement) -> Frame:
    """
    Convert an IFC LocalPlacement to a COMPAS frame.
    """

    Location = placement.RelativePlacement.Location
    Axis = placement.RelativePlacement.Axis
    RefDirection = placement.RelativePlacement.RefDirection

    if Axis and RefDirection:
        zaxis = Vector(*Axis.DirectionRatios)
        xaxis = Vector(*RefDirection.DirectionRatios)
        yaxis = zaxis.cross(xaxis)
        xaxis = yaxis.cross(zaxis)
    else:
        xaxis = Vector.Xaxis()
        yaxis = Vector.Yaxis()

    point = Point(*Location.Coordinates)
    return Frame(point, xaxis, yaxis)


def IfcGridPlacement_to_transformation(placement) -> Transformation:
    pass



from typing import List

from compas_ifc.entities.buildingelements import BuildingElement
from compas_ifc.entities.buildingstorey import BuildingStorey
from compas_ifc.entities.space import Space
from compas_ifc.entities.spatialelement import SpatialElement


class Building(SpatialElement):
    """
    Class representing an IFC building.

    Attributes
    ----------
    address : str
        The compiled address of the building.
    building_storeys : List[BuildingStorey]
        The building_storeys contained in the building.
        If the building_storeys are not properly nested, they can be found at the top level of an IFC model.
    spaces : List[Space]
        The spaces contained in the building.
    building_elements : List[BuildingElement]
        The building elements contained in the building.

    Notes
    -----
    The order of spatial structure elements being included in the concept for builing projects are from high to low level:
    IfcProject, IfcSite, IfcBuilding, IfcBuildingStorey, and IfcSpace with IfcSite, IfcBuildingStorey and IfcSpace being optional levels.
    Therefore an spatial structure element can only be part of an element at the same or higher level [1]_.

    This means the minimum spatial structure of a project is:

    * IfcProject

      * IfcBuilding

        * IfcBuildingElement

    Therefore, all building elements found in a model can be accessed through the Building class.

    .. code-block:: python

        model = Model(filepath)
        for project in model.projects:
            if project.sites:
                for site in project.sites:
                    for building in site.buildings:
                        for element in building.elements:
                            # ...
            else:
                for building in project.buildings:
                    for element in building.elements:
                        # ...

    References
    ----------
    .. [1] :ifc:`spatial-composition`

    """

    def __init__(self, entity, model):
        super().__init__(entity, model)
        self._building_storeys = None
        self._spaces = None
        self._elements = None

    @property
    def building_storeys(self) -> List[BuildingStorey]:
        return [entity for entity in self.traverse() if entity.is_a("IfcBuildingStorey")]

    @property
    def spaces(self) -> List[Space]:
        return [entity for entity in self.traverse() if entity.is_a("IfcSpace")]

    @property
    def building_elements(self) -> List[BuildingElement]:
        return [entity for entity in self.traverse() if entity.is_a("IfcBuildingElement")]

    @property
    def address(self) -> str:
        attr = self.attribute("BuildingAddress")
        address = ""
        address += ", ".join(attr["AddressLines"])
        address += f", {attr['Country']}-{attr['PostalCode']}"
        address += f" {attr['Town']}"
        return address



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from compas_ifc.entities.generated.IFC4 import IfcSite
else:
    IfcSite = object


class IfcSite(IfcSite):
    @property
    def buildings(self):
        return self.children_by_type("IfcBuilding", recursive=True)

    @property
    def building_elements(self):
        return self.children_by_type("IfcBuildingElement", recursive=True)

    @property
    def geographic_elements(self):
        return self.children_by_type("IfcGeographicElement", recursive=True)



import ifcopenshell

# TODO: pre-generate all classes.
# TODO: separate ifcopenshell APIs


class Factory:
    EXTENSIONS = {}

    def __init__(self, schema: str = "IFC4") -> None:
        self.schema = ifcopenshell.ifcopenshell_wrapper.schema_by_name(schema)

    def get_extensions(self, name, declaration):
        name = declaration.name()
        inheritances = [name]
        supertype = declaration.supertype()
        while supertype:
            name = supertype.name()
            inheritances.append(name)
            supertype = supertype.supertype()

        inheritances = inheritances[::-1]
        extensions = []
        for cls in inheritances:
            if cls in self.EXTENSIONS:
                extensions.append(self.EXTENSIONS[cls])
        return tuple(extensions)

    def create_class(self, name):
        declaration = self.schema.declaration_by_name(name)
        extensions = self.get_extensions(name, declaration)

        def init(self, entity):
            self._entity = entity

        attributes = {
            "__init__": init,
            "__repr__": lambda self: f"<compas_ifc.{self._entity.is_a()}>",
        }

        for attribute in declaration.all_attributes():
            name = attribute.name()
            attributes[name] = self.create_getter(attribute)

        for attribute in declaration.all_inverse_attributes():
            name = attribute.name()
            attributes[name] = self.create_getter(attribute, setter=False)

        return type(name, extensions, attributes)

    def create_getter(self, attribute, setter=True):
        name = attribute.name()

        @property
        def prop(self):
            return getattr(self._entity, name)

        if setter:

            @prop.setter
            def prop(self, value):
                if not hasattr(self._entity, name):
                    raise AttributeError(f"Attribute <{name}> does not exist")

                # Make checks pluggable functions

                # check aggregation type
                attribute_type = attribute.type_of_attribute()
                if attribute_type.as_aggregation_type():
                    value = list(value)
                    lower = attribute_type.bound1()
                    upper = attribute_type.bound2()
                    if lower > 0 and upper > 0:
                        if len(value) < lower or len(value) > upper:
                            raise ValueError(f"Expected <{name}> to have between {lower} and {upper} elements")
                    # TODO: check type of elements

                # check entity type
                if attribute_type.as_named_type():
                    declared_type = attribute_type.declared_type()
                    type_name = declared_type.name()
                    # TODO: take care of non-entity situations
                    if declared_type.as_entity() and value.is_a() != type_name:
                        raise TypeError(f"Expected <{name}> to be of type {type_name}, got {value.is_a()}")

                setattr(self._entity, name, value)

        return prop


if __name__ == "__main__":
    file = ifcopenshell.open("data/wall-with-opening-and-window.ifc")
    _entity = file[34]
    declaration = _entity.is_a()

    from extensions import EXTENSIONS

    Factory.EXTENSIONS = EXTENSIONS
    print(Factory.EXTENSIONS)

    factory = Factory()
    Entity = factory.create_class(declaration)

    entity = Entity(_entity)
    entity.test()
    entity.test2()
    print(entity)

    # print(entity.ThePerson)
    # # entity.ThePerson = "some wrong value type"
    # entity.ThePerson = file[3]

    # print(entity.ThePerson)
    # print(entity)



import ifcopenshell
from compas.datastructures import Mesh


def mesh_to_IfcPolygonalFaceSet(file: ifcopenshell.file, mesh: Mesh) -> ifcopenshell.entity_instance:
    """
    Convert a COMPAS mesh to an IFC PolygonalFaceSet.
    """
    _vertices = []
    for key, attr in mesh.vertices(True):
        _vertices.append((key, (float(attr["x"]), float(attr["y"]), float(attr["z"]))))
    _vertices = sorted(_vertices, key=lambda x: x[0])
    vertices = [v[1] for v in _vertices]

    keys = [v[0] for v in _vertices]

    faces = []
    for fkey in mesh.faces():
        indexes = [keys.index(i) + 1 for i in mesh.face_vertices(fkey)]
        faces.append(file.createIfcIndexedPolygonalFace(indexes))

    return file.create_entity(
        "IfcPolygonalFaceSet",
        Coordinates=file.createIfcCartesianPointList3D(vertices),
        Faces=faces,
    )


def mesh_to_IfcShellBasedSurfaceModel(file: ifcopenshell.file, mesh: Mesh) -> ifcopenshell.entity_instance:
    """
    Convert a COMPAS mesh to an IFC PolygonalFaceSet.
    """
    vertices = {}
    for key, attr in mesh.vertices(True):
        vertex = file.createIfcCartesianPoint((float(attr["x"]), float(attr["y"]), float(attr["z"])))
        vertices[key] = vertex

    faces = []
    for fkey in mesh.faces():
        indexes = [vertices[key] for key in mesh.face_vertices(fkey)]
        polyloop = file.create_entity("IfcPolyLoop", indexes)
        bound = file.create_entity("IfcFaceOuterBound", polyloop, True)
        face = file.create_entity("IfcFace", [bound])
        faces.append(face)

    shell = file.create_entity("IfcOpenShell", faces)
    return file.create_entity("IfcShellBasedSurfaceModel", [shell])



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from compas_ifc.entities.generated.IFC4 import IfcSpatialElement
else:
    IfcSpatialElement = object


class IfcSpatialElement(IfcSpatialElement):
    @property
    def children(self):
        children = super().children
        children += sum([relation.RelatedElements for relation in self.ContainsElements()], [])
        # NOTE: in IFC2X3 this is in IfcSpatialStructureElement
        return list(set(children))



import inspect
import os
import types

import ifcopenshell

from compas_ifc.entities import extensions


class Generator:
    def __init__(self, schema="IFC4"):
        self.schema = ifcopenshell.ifcopenshell_wrapper.schema_by_name(schema)

    def generate(self):
        HERE = os.path.dirname(__file__)
        FOLDER = os.path.join(HERE, "generated", self.schema.name())
        if not os.path.exists(FOLDER):
            os.makedirs(FOLDER)

        doc_string = """
.. autosummary::
    :toctree: generated/
    :nosignatures:
    :template: class.rst

"""
        init_string = ""

        count = 0

        for declaration in self.schema.declarations():
            class_string = None
            if declaration.as_entity():
                entity_generator = EntityGenerator(declaration)
                class_string = entity_generator.generate()
                name = entity_generator.name

            # if declaration.as_type_declaration():
            #     type_declaration_generator = TypeDeclarationGenerator(declaration)
            #     class_string = type_declaration_generator.generate()
            #     name = type_declaration_generator.name

            if declaration.as_enumeration_type():
                enum_generator = EnumGenerator(declaration)
                class_string = enum_generator.generate()
                name = enum_generator.name

            if class_string:
                init_string += f"from .{name.lower()} import {name}\n"
                doc_string += f"    {name}\n"

                with open(os.path.join(FOLDER, f"{name.lower()}.py"), "w") as f:
                    f.write(class_string)
                    count += 1

        init_string = f'"""{doc_string}"""\n\n{init_string}'

        with open(os.path.join(FOLDER, "__init__.py"), "w") as f:
            f.write(init_string)

        print(f"Generated {count} classes for at {FOLDER}.")


class EntityGenerator:
    TEMPLATE = """IMPORTS
class CLASS_NAME(PARENT_NAME):
    \"\"\"DESCRIPTION\"\"\"
"""

    def __init__(self, declaration):
        self.declaration = declaration
        self.name = declaration.name()
        self.imports = set()
        self.attribute_imports = set()
        self.parent = None
        self.extension = None
        self.attributes = []
        self.inverse_attributes = []
        self.description = ""

    def get_parent(self):
        if self.declaration.supertype():
            self.parent = self.declaration.supertype().name()
            self.imports.add(f"from .{self.parent.lower()} import {self.parent}")
        else:
            self.parent = "Base"
            self.imports.add("from compas_ifc.entities.base import Base")

        extension = getattr(extensions, self.name, None)
        if extension:
            print("Found extension class:", extension)
            self.imports.add(f"from compas_ifc.entities.extensions import {self.name} as {self.name}_Ext   # type: ignore")
            self.extension = f"{self.name}_Ext"

    def get_description(self):
        self.description = f"Wrapper class for {self.name}."

    def get_attributes(self):
        # attributes = self.declaration.attributes()
        # direved = self.declaration.derived()
        # all_attributes = self.declaration.all_attributes()
        # all_inverse_attributes = self.declaration.all_inverse_attributes()

        derived = self.declaration.derived()
        attribute_names = [attr.name() for attr in self.declaration.attributes()]

        for i, attribute in enumerate(self.declaration.all_attributes()):
            if attribute.name() in attribute_names:
                self.attributes.append(AtrtributeGenerator(attribute, self, False))
            elif derived[i]:
                self.attributes.append(AtrtributeGenerator(attribute, self, True))

        inverse_attributes_from_supertype = []
        if self.declaration.supertype():
            for ia in self.declaration.supertype().all_inverse_attributes():
                inverse_attributes_from_supertype.append(ia.name())

        for inverse_attribute in self.declaration.all_inverse_attributes():
            if inverse_attribute.name() not in inverse_attributes_from_supertype:
                self.inverse_attributes.append(InverseAtrtributeGenerator(inverse_attribute, self))

    def get_attribute_imports_string(self):
        if not self.attribute_imports:
            return ""

        attribute_imports_string = "from typing import TYPE_CHECKING\nif TYPE_CHECKING:\n"
        for import_string in sorted(self.attribute_imports):
            attribute_imports_string += f"    {import_string}\n"
        return attribute_imports_string

    def get_extension_methods(self, class_name):
        extension = getattr(extensions, class_name, None)
        if extension is None:
            return ""

        print("Extension classes:", extension)

        extension_methods_string = ""
        for method in dir(extension):
            method = getattr(extension, method)
            if isinstance(method, types.FunctionType):
                extension_methods_string += f"\n{inspect.getsource(method)}"
            elif isinstance(method, property):
                if method.fget is not None:
                    extension_methods_string += f"\n{inspect.getsource(method.fget)}"
                if method.fset is not None:
                    extension_methods_string += f"\n{inspect.getsource(method.fset)}"

        return extension_methods_string

    def generate(self):
        self.get_parent()
        self.get_description()
        self.get_attributes()

        class_string = self.TEMPLATE.replace("CLASS_NAME", self.name)
        if self.extension:
            class_string = class_string.replace("PARENT_NAME", f"{self.extension}, {self.parent}")
        else:
            class_string = class_string.replace("PARENT_NAME", self.parent)
        class_string = class_string.replace("DESCRIPTION", self.description)

        for attribute in self.attributes:
            class_string += attribute.generate()
            self.attribute_imports.update(attribute.imports)

        for inverse_attribute in self.inverse_attributes:
            class_string += inverse_attribute.generate()
            self.attribute_imports.update(inverse_attribute.imports)

        if class_string.find("Union") != -1:
            self.imports.add("from typing import Union")

        import_strings = "\n".join(sorted(self.imports)) + "\n\n" + self.get_attribute_imports_string()

        class_string = class_string.replace("IMPORTS", import_strings)

        # class_string += self.get_extension_methods(self.name)

        return class_string


class AtrtributeGenerator:
    TEMPLATE = """
    @property
    def ATTRIBUTE_NAME(self)-> ATTRIBUTE_TYPE:
        \"\"\"DESCRIPTION\"\"\"
        return self._get_attribute("ATTRIBUTE_NAME")

    @ATTRIBUTE_NAME.setter
    def ATTRIBUTE_NAME(self, value: ATTRIBUTE_TYPE):
        return self._set_attribute("ATTRIBUTE_NAME", value)
"""
    TEMPLATE_DERIVED = """
    @property
    def ATTRIBUTE_NAME(self)-> ATTRIBUTE_TYPE:
        \"\"\"DESCRIPTION\"\"\"
        return self._get_attribute("ATTRIBUTE_NAME")

    @ATTRIBUTE_NAME.setter
    def ATTRIBUTE_NAME(self, value: ATTRIBUTE_TYPE):
        # Derived attribute
        pass
"""

    TYPE_MAP = {
        "DOUBLE": "float",
        "INT": "int",
        "STRING": "str",
        "LOGICAL": "bool",
        "BOOL": "bool",
        "BINARY": "bytes",
    }

    def __init__(self, attribute, parent, is_derived):
        self.parent = parent
        self.attribute = attribute
        self.is_derived = is_derived
        self.name = attribute.name()
        self.imports = set()
        self.type = None
        self.description = str(attribute)

    def get_aggregation_type(self, attribute_type):
        # type_aggragation = attribute_type.type_of_aggregation()
        # bound1 = attribute_type.bound1()
        # bound2 = attribute_type.bound2()
        # TODO: add support for bounds etc.

        type_of_element = attribute_type.type_of_element()
        if type_of_element.as_aggregation_type():
            aggregation_string = self.get_aggregation_type(type_of_element)
            return f"list[{aggregation_string}]"
        else:
            declared_type = type_of_element.declared_type()
            if isinstance(declared_type, str):
                # TODO: this duplicates the code
                TYPE_MAP = {
                    "integer": "int",
                    "logical": "bool",
                    "boolean": "bool",
                    "real": "float",
                    "binary": "bytes",
                }
                type_of_element_string = TYPE_MAP[declared_type]
            else:
                if type_of_element.declared_type().as_select_type():
                    type_of_element_string = self.get_select_type(type_of_element)
                elif type_of_element.declared_type().as_type_declaration():
                    type_of_element_string = self.get_type_declaration(type_of_element)
                else:
                    type_of_element_string = type_of_element.declared_type().name()
                    if type_of_element_string != self.parent.name:
                        self.imports.add(f"from .{type_of_element_string.lower()} import {type_of_element_string}")
                    type_of_element_string = f'"{type_of_element_string}"'
            return f"list[{type_of_element_string}]"

    def get_select_type(self, attribute_type):
        def flaten_select_list(select_type, initial_list=None):
            if initial_list is None:
                initial_list = []
            select_list = select_type.select_list()
            for item in select_list:
                if item.as_select_type():
                    initial_list = flaten_select_list(item, initial_list)
                else:
                    initial_list.append(item)
            return initial_list

        select_list = flaten_select_list(attribute_type.declared_type())
        class_names = set()
        for item in select_list:
            if item.as_type_declaration():
                class_names.add(self.get_type_declaration(item))
            else:
                class_names.add(f'"{item.name()}"')
                if item.name() != self.parent.name:
                    self.imports.add(f"from .{item.name().lower()} import {item.name()}")
        select_string = "Union[" + ", ".join(class_names) + "]"
        return select_string

    def get_type_declaration(self, attribute_type):
        if hasattr(attribute_type, "argument_types"):
            declared_type = attribute_type
        else:
            declared_type = attribute_type.declared_type()

        ifc_type = declared_type.argument_types()[0]
        if ifc_type.startswith("AGGREGATE OF"):
            value_type = ifc_type.split("OF ")[1]
            if value_type == "ENTITY INSTANCE":
                python_type = "list"  # TODO: handle this
            else:
                python_type = f"list[{self.TYPE_MAP[value_type]}]"
        else:
            python_type = self.TYPE_MAP[ifc_type]
        return python_type

    def get_attribute_type(self):
        attribute_type = self.attribute.type_of_attribute()
        if attribute_type.as_aggregation_type():
            self.type = self.get_aggregation_type(attribute_type)
        elif attribute_type.as_simple_type():
            attribute_type = attribute_type.declared_type()
            TYPE_MAP = {
                "integer": "int",
                "logical": "bool",
                "boolean": "bool",
                "real": "float",
                "string": "str",
            }
            self.type = TYPE_MAP[attribute_type]
        elif attribute_type.declared_type().as_select_type():
            self.type = self.get_select_type(attribute_type)
        elif attribute_type.declared_type().as_type_declaration():
            self.type = self.get_type_declaration(attribute_type)
        else:
            # Entity, Enumeration
            type_name = attribute_type.declared_type().name()
            if type_name != self.parent.name:
                self.imports.add(f"from .{type_name.lower()} import {type_name}")
            self.type = f'"{type_name}"'

    def generate(self):
        self.get_attribute_type()
        if self.is_derived:
            attribute_string = self.TEMPLATE_DERIVED.replace("ATTRIBUTE_NAME", self.name)
        else:
            attribute_string = self.TEMPLATE.replace("ATTRIBUTE_NAME", self.name)
        attribute_string = attribute_string.replace("ATTRIBUTE_TYPE", str(self.type))
        attribute_string = attribute_string.replace("DESCRIPTION", self.description)

        return attribute_string


class InverseAtrtributeGenerator(AtrtributeGenerator):
    TEMPLATE = """
    def ATTRIBUTE_NAME(self)-> list[ATTRIBUTE_TYPE]:
        \"\"\"DESCRIPTION\"\"\"
        return self._get_inverse_attribute("ATTRIBUTE_NAME")
"""

    def __init__(self, attribute, parent):
        self.parent = parent
        self.attribute = attribute
        self.name = attribute.name()
        self.imports = set()
        self.type = None
        self.description = str(attribute)

    def get_attribute_type(self):
        # TODO: handle bound1, bound2 etc. for the aggregation
        entity_reference = self.attribute.entity_reference()
        self.type = entity_reference.name()
        self.imports.add(f"from .{self.type.lower()} import {self.type}")

    def generate(self):
        self.get_attribute_type()
        attribute_string = self.TEMPLATE.replace("ATTRIBUTE_NAME", self.name)
        attribute_string = attribute_string.replace("ATTRIBUTE_TYPE", f'"{self.type}"')
        attribute_string = attribute_string.replace("DESCRIPTION", self.description)

        return attribute_string


class TypeDeclarationGenerator:
    TEMPLATE = """class CLASS_NAME(PARENT_NAME):
    \"\"\"Some description.\"\"\"
"""

    TYPE_MAP = {
        "DOUBLE": "float",
        "INT": "int",
        "STRING": "str",
        "LOGICAL": "bool",
        "BOOL": "bool",
        "BINARY": "bytes",
    }

    def __init__(self, declaration):
        self.declaration = declaration
        self.name = declaration.name()
        self.parent = None

    def get_parent(self):
        ifc_type = self.declaration.argument_types()[0]
        if ifc_type.startswith("AGGREGATE OF"):
            value_type = ifc_type.split("OF ")[1]
            if value_type == "ENTITY INSTANCE":
                python_type = "list"  # TODO: handle this
            else:
                python_type = f"list[{self.TYPE_MAP[value_type]}]"
        else:
            python_type = self.TYPE_MAP[ifc_type]
        self.parent = python_type

    def generate(self):
        self.get_parent()

        class_string = self.TEMPLATE.replace("CLASS_NAME", self.name)
        class_string = class_string.replace("PARENT_NAME", self.parent)

        return class_string


class EnumGenerator:
    TEMPLATE = """class CLASS_NAME(str):
    items = ITEMS
"""

    def __init__(self, declaration):
        self.declaration = declaration
        self.name = declaration.name()
        self.items = []

    def get_items(self):
        for item in self.declaration.enumeration_items():
            self.items.append(item)

    def generate(self):
        self.get_items()

        class_string = self.TEMPLATE.replace("CLASS_NAME", self.name)
        class_string = class_string.replace("ITEMS", str(self.items))

        return class_string


if __name__ == "__main__":
    generator = Generator(schema="IFC2X3")
    generator.generate()

    generator = Generator(schema="IFC4")
    generator.generate()



# from typing import Iterator
from typing import Dict
from typing import List

from compas.geometry import Frame
from compas.geometry import Vector

from compas_ifc.entities.objectdefinition import ObjectDefinition
from compas_ifc.entities.site import Building
from compas_ifc.entities.site import Site
from compas_ifc.resources import IfcAxis2Placement3D_to_frame


class Project(ObjectDefinition):
    """
    Class representing IFC projects.

    Attributes
    ----------
    sites : List[:class:`Site`]
        The sites contained in the project.
    buildings : List[:class:`Building`]
        The buildings contained in the project.
        Note that this list is empty if ``sites`` is not empty.
    units : List[Dict]
        The SI units used in the project.
    contexts : List[Dict]
        The representation contexts included in the project.
        Each representation context defines a coordinates system, a "true north" vector, and numerical precision.
    frame : :class:`compas.geometry.Frame` or None
        The reference frame, defined by the "Model" context, if that context exists.
    north : :class:`compas.geometry.Vector` or None
        The north vector with respect to the reference frame.
    gps : ???
        Positioning of the reference frame on earth.

    Notes
    -----
    The order of spatial structure elements being included in the concept for builing projects are from high to low level:
    IfcProject, IfcSite, IfcBuilding, IfcBuildingStorey, and IfcSpace with IfcSite, IfcBuildingStorey and IfcSpace being optional levels.
    Therefore an spatial structure element can only be part of an element at the same or higher level [2]_.

    This means the minimum spatial structure of a project is:

    * IfcProject
      * IfcBuilding

    Therefore, a project will contain one or more sites, or one or more buildings.
    If the project defines sites, the buildings will be contained in the sites.

    .. code-block:: python

        model = Model(filepath)
        for project in model.projects:
            if not project.sites:
                for building in project.buildings:
                    # ...
            else:
                for site in project.sites:
                    for building in site.buildings:
                        # ...

    References
    ----------
    .. [1] :ifc:`project-context`
    .. [2] :ifc:`spatial-composition`

    """

    def __init__(self, entity, model):
        super().__init__(entity, model)
        self._units = None
        self._contexts = None
        self._sites = None
        self._buildings = None
        self._length_unit = {"type": "LENGTHUNIT", "name": "METRE", "prefix": "MILLI"}
        # TODO: deal with other units

    @property
    def sites(self) -> List[Site]:
        return [entity for entity in self.children if entity.is_a("IfcSite")]

    @property
    def buildings(self) -> List[Building]:
        return [entity for entity in self.traverse() if entity.is_a("IfcBuilding")]

    @property
    def contexts(self) -> List[Dict]:
        if self._contexts is None:
            self._contexts = []
            for context in self._entity.RepresentationContexts:
                north = Vector(*context.TrueNorth.DirectionRatios)
                wcs = IfcAxis2Placement3D_to_frame(context.WorldCoordinateSystem)
                self._contexts.append(
                    {
                        "identifier": context.ContextIdentifier,
                        "type": context.ContextType,
                        "precision": context.Precision,
                        "dimension": context.CoordinateSpaceDimension,
                        "north": north,
                        "wcs": wcs,
                    }
                )
        return self._contexts

    @property
    def units(self) -> List[Dict]:
        if self._units is None:
            self._units = []
            for unit in self._entity.UnitsInContext.Units:
                if unit.is_a("IfcSIUnit"):
                    self._units.append(
                        {
                            "type": unit.UnitType,
                            "name": unit.Name,
                            "prefix": unit.Prefix,
                        }
                    )
        return self._units

    @property
    def length_unit(self):
        if self._entity:
            for unit in self.units:
                if unit["type"] == "LENGTHUNIT":
                    return unit
        else:
            return self._length_unit

    @length_unit.setter
    def length_unit(self, unit):
        self._length_unit = unit

    @property
    def length_scale(self):
        unit = self.length_unit
        if unit:
            if unit["name"] == "METRE" and not unit["prefix"]:
                return 1.0
            if unit["name"] == "METRE" and unit["prefix"] == "CENTI":
                return 1e-2
            if unit["name"] == "METRE" and unit["prefix"] == "MILLI":
                return 1e-3
        return 1.0

    @property
    def frame(self) -> Frame:
        if self._entity:
            for context in self.contexts:
                if context["type"] == "Model":
                    return context["wcs"]

    @property
    def north(self) -> Vector:
        for context in self.contexts:
            if context["type"] == "Model":
                return context["north"]

    @property
    def gps(self):
        # global positioning of the reference frame on earth
        pass

    # @property
    # def owner(self):

    def add_site(self, site: Site) -> None:
        """
        Add a site to the project.

        Parameters
        ----------
        site : :class:`Site`

        Returns
        -------
        None

        """
        raise NotImplementedError



"""
********************************************************************************
compas_ifc
********************************************************************************

.. currentmodule:: compas_ifc


.. toctree::
    :maxdepth: 1

    compas_ifc.attributes
    compas_ifc.entities
    compas_ifc.helpers
    compas_ifc.model
    compas_ifc.representation
    compas_ifc.resources
    compas_ifc.viewer

"""

from __future__ import print_function

import os


__author__ = ["tom van mele"]
__copyright__ = "ETH Zurich"
__license__ = "MIT License"
__email__ = "van.mele@arch.ethz.ch"
__version__ = "1.2.1"


HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, "../../"))
DATA = os.path.abspath(os.path.join(HOME, "data"))
DOCS = os.path.abspath(os.path.join(HOME, "docs"))
TEMP = os.path.abspath(os.path.join(HOME, "temp"))


__all__ = ["HOME", "DATA", "DOCS", "TEMP"]

__all_plugins__ = [
    "compas_ifc.brep",
]



from typing import List

import ifcopenshell

# from compas.geometry import Polyline
from compas.geometry import Box
from compas.geometry import Line
from compas.geometry import Point
from compas.geometry import Transformation

from compas_ifc.resources.geometry import IfcAxis2Placement3D_to_frame
from compas_ifc.resources.geometry import IfcDirection_to_vector
from compas_ifc.resources.geometry import IfcProfileDef_to_curve


def IfcAdvancedBrep_to_brep(advanced_brep):
    """
    Convert an IFC AdvancedBrep [advancedbrep]_ to a COMPAS brep.

    """
    pass


def IfcAdvancedBrepWithVoids_to_brep(advanced_brep_with_voids):
    """
    Convert an IFC AdvancedBrepWithVoids [advancedbrepwithvoids]_ to a COMPAS brep.

    """
    pass


def IfcBlock_to_box(block) -> Box:
    """
    Convert an IFC Block [block]_ to a COMPAS box.

    """
    pass


def IfcBooleanClippingResult_to_brep(boolean_clipping_result):
    """
    Convert an IFC BooleanClippingResult [booleanclippingresult]_ to a COMPAS brep.

    """
    return IfcBooleanResult_to_brep(boolean_clipping_result)


def IfcBooleanResult_to_brep(boolean_result):
    """
    Convert an IFC BooleanResult [booleanresult]_ to a COMPAS brep.

    """
    from compas_occ.brep import OCCBrep

    br = boolean_result

    # First Operand
    if br.FirstOperand.is_a("IfcBooleanResult"):
        A = IfcBooleanResult_to_brep(br.FirstOperand)
    elif br.FirstOperand.is_a("IfcExtrudedAreaSolid"):
        A = IfcExtrudedAreaSolid_to_brep(br.FirstOperand)
    else:
        raise NotImplementedError

    # Second Operand
    if br.SecondOperand.is_a("IfcBoxedHalfSpace"):
        B = IfcBoxedHalfSpace(br.SecondOperand)
    elif br.SecondOperand.is_a("IfcPolygonalBoundedHalfSpace"):
        B = IfcPolygonalBoundedHalfSpace_to_brep(br.SecondOperand)
    elif br.SecondOperand.is_a("IfcPolygonalFaceSet"):
        B = IfcPolygonalFaceSet_to_brep(br.SecondOperand)
    else:
        raise NotImplementedError(br.SecondOperand)

    # Operator
    if br.Operator == "UNION":
        C: OCCBrep = A + B
    elif br.Operator == "INTERSECTION":
        C: OCCBrep = A & B
    elif br.Operator == "DIFFERENCE":
        C: OCCBrep = A - B
    else:
        raise NotImplementedError(br.Operator)

    C.sew()
    C.fix()
    C.make_solid()
    return C


def IfcIndexedPolyCurve_to_lines(axis) -> List[Line]:
    """
    Convert an IFC Axis [axis]_ to a COMPAS polyline.

    """
    lines = []
    points = IfcCartesianPointList(axis.Points)
    for segment in axis.Segments:
        start, end = segment.wrappedValue
        start -= 1
        end -= 1
        lines.append(Line(points[start], points[end]))
    return lines


def IfcBoundingBox_to_box(bounding_box) -> Box:
    """
    Convert an IFC BoundingBox [boundingbox]_ to a COMPAS box.

    """
    a = Point(*bounding_box.Corner.Coordinates)
    b = a + [bounding_box.XDim, bounding_box.YDim, 0]
    box = Box.from_corner_corner_height(a, b, bounding_box.ZDim)
    return box


def IfcBoxedHalfSpace(boxed_half_space):
    bhs = boxed_half_space
    print(bhs.BaseSurface)
    print(bhs.Enclosure)
    print(bhs.AgreementFlag)


def IfcCartesianPointList(cartesian_point_list) -> List[Point]:
    return [Point(*p) for p in cartesian_point_list.CoordList]


def IfcExtrudedAreaSolid_to_brep(extruded_area_solid):
    from compas_occ.brep import OCCBrep

    eas = extruded_area_solid
    profile = IfcProfileDef_to_curve(eas.SweptArea)

    position = IfcAxis2Placement3D_to_frame(eas.Position)
    transformation = Transformation.from_frame(position)

    def _extrude(profile, eas):
        vector = IfcDirection_to_vector(eas.ExtrudedDirection)
        vector.scale(eas.Depth)
        brep = OCCBrep.from_extrusion(profile, vector)
        brep.sew()
        brep.fix()
        brep.make_solid()
        brep.transform(transformation)
        return brep

    if isinstance(profile, list):
        extrusions = [_extrude(p, eas) for p in profile]
        union = extrusions[0]
        for e in extrusions[1:]:
            union = union + e
        union.sew()
        union.fix()
        union.make_solid()
        return union
    else:
        return _extrude(profile, eas)


def IfcFacetedBrep_to_brep(faceted_brep):
    pass


def IfcFacetedBrepWithVoids_to_brep(faceted_brep_with_voids):
    # fbwv = faceted_brep_with_voids
    pass


def IfcIndexedPolygonalFaceSet_to_brep():
    pass


def IfcPolygonalBoundedHalfSpace_to_brep(polygonal_bounded_half_space):
    pbhs = polygonal_bounded_half_space
    print(pbhs.Position)
    print(pbhs.PolygonalBoundary)
    print(pbhs.BaseSurface)


def IfcPolygonalFaceSet_to_brep(polygonal_face_set):
    from compas_occ.brep import OCCBrep

    pfs = polygonal_face_set

    xyz = pfs.Coordinates.CoordList
    vertices = [[float(x), float(y), float(z)] for x, y, z in xyz]
    faces = [[i - 1 for i in face.CoordIndex] for face in pfs.Faces]
    polygons = [[vertices[index] for index in face] for face in faces]
    brep = OCCBrep.from_polygons(polygons)

    brep.sew()
    brep.fix()
    brep.make_solid()

    return brep


def IfcTessellatedFaceSet_to_brep(tessellated_face_set):
    tfs = tessellated_face_set

    if tfs.is_a("IfcTriangulatedFaceSet"):
        return IfcTriangulatedFaceSet_to_brep(tfs)

    if tfs.is_a("IfcPolygonalFaceSet"):
        return IfcPolygonalFaceSet_to_brep(tfs)

    raise NotImplementedError(tfs.is_a())


def IfcTriangulatedFaceSet_to_brep(triangulated_face_set):
    from compas_occ.brep import OCCBrep

    tfs = triangulated_face_set

    xyz = tfs.Coordinates.CoordList
    vertices = [[float(x), float(y), float(z)] for x, y, z in xyz]
    faces = [[a - 1, b - 1, c - 1] for a, b, c in tfs.CoordIndex]
    triangles = [[vertices[index] for index in face] for face in faces]
    brep = OCCBrep.from_polygons(triangles)

    brep.sew()
    brep.fix()
    brep.make_solid()

    return brep


def IfcShape_to_brep(ifc_shape):
    from compas_occ.brep import OCCBrep

    settings = ifcopenshell.geom.settings()
    settings.set(settings.USE_PYTHON_OPENCASCADE, True)
    shape = ifcopenshell.geom.create_shape(settings, ifc_shape)

    brep = OCCBrep.from_shape(shape)
    brep.sew()
    brep.fix()
    brep.make_solid()

    return brep


def IfcShape_to_tessellatedbrep(ifc_shape):
    from compas_ifc.brep import TessellatedBrep

    settings = ifcopenshell.geom.settings()
    shape = ifcopenshell.geom.create_shape(settings, ifc_shape)
    return TessellatedBrep(vertices=shape.verts, edges=shape.edges, faces=shape.faces)


# # IfcTessellationItem_to_mesh
# def tessellation_to_mesh(item) -> Mesh:
#     """
#     Convert a tessellation item to a Mesh.

#     """
#     mesh = Mesh()
#     mesh.update_default_face_attributes(colour=None)

#     face_color = {}

#     if item.HasColours:
#         colourmap = item.HasColours[0]
#         for index, face in zip(
#             colourmap.ColourIndex,
#             colourmap.MappedTo.Faces,
#         ):
#             fid = face.id()
#             colour = colourmap.Colours.ColourList[index - 1]
#             face_color[fid] = Color(*colour)

#     for face in item.Faces:
#         vertices = []
#         for index in face.CoordIndex:
#             x, y, z = item.Coordinates.CoordList[index - 1]
#             vertices.append(mesh.add_vertex(x=x, y=y, z=z))
#         mesh.add_face(vertices, colour=face_color.get(face.id()))
#     return mesh



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from compas_ifc.entities.generated.IFC2X3 import IfcSpatialStructureElement
else:
    IfcSpatialStructureElement = object


class IfcSpatialStructureElement(IfcSpatialStructureElement):
    @property
    def children(self):
        children = super().children
        children += sum([relation.RelatedElements for relation in self.ContainsElements()], [])
        # NOTE: in IFC4 this is in IfcSpatialElement
        return list(set(children))



import ifcopenshell
from compas.geometry import Frame
from compas.geometry import Point

from .shapes import create_IfcAxis2Placement3D


def point_to_ifc_cartesian_point(file: ifcopenshell.file, point: Point) -> ifcopenshell.entity_instance:
    """
    Convert a COMPAS point to an IFC Cartesian Point.
    """
    point = [float(i) for i in point]
    return file.create_entity("IFCCARTESIANPOINT", (*point,))


def occ_plane_to_frame(occ_plane) -> Frame:
    """
    Convert an OCC plane to a COMPAS frame.
    """
    location = occ_plane.Location().Coord()
    xdir = occ_plane.XAxis().Direction().Coord()
    ydir = occ_plane.YAxis().Direction().Coord()
    return Frame(location, xdir, ydir)


def frame_to_ifc_plane(file: ifcopenshell.file, frame: Frame) -> ifcopenshell.entity_instance:
    """
    Convert a COMPAS frame to an IFC Plane.
    """
    IfcAxis2Placement3D = create_IfcAxis2Placement3D(file, frame.point, frame.zaxis, frame.xaxis)
    return file.create_entity("IfcPlane", IfcAxis2Placement3D)



"""
********************************************************************************
resources
********************************************************************************

.. currentmodule:: compas_ifc.resources

This module contains functions for converting between IFC geometry resources and equivalent COMPAS geometry objects.
For more information about IFC geometric entities see [geometricmodelresource]_, [geometryresource]_ and [geometricconstraintresource]_.


Functions
=========

Geometry
--------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    IfcCartesianPoint_to_point
    IfcDirection_to_vector
    IfcVector_to_vector
    IfcLine_to_line
    IfcPlane_to_plane
    IfcAxis2Placement3D_to_frame
    IfcCartesianTransformationOperator3D_to_frame


Geometric Constraint
--------------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    IfcGridPlacement_to_transformation
    IfcLocalPlacement_to_transformation


Geometric Model
---------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    IfcAdvancedBrep_to_brep
    IfcAdvancedBrepWithVoids_to_brep
    IfcBlock_to_box
    IfcBooleanClippingResult_to_brep
    IfcBoundingBox_to_box
    IfcExtrudedAreaSolid_to_brep
    IfcFacetedBrep_to_brep
    IfcFacetedBrepWithVoids_to_brep
    IfcIndexedPolygonalFaceSet_to_brep
    IfcPolygonalBoundedHalfSpace_to_brep
    IfcPolygonalFaceSet_to_brep
    IfcTriangulatedFaceSet_to_brep


References
==========

.. [geometricmodelresource] :ifc:`ifcgeometricmodelresource`
.. [geometryresource] :ifc:`ifcgeometryresource`
.. [geometricconstraintresource] :ifc:`ifcgeometricconstraintresource`

.. [cartesianpoint] :ifc:`ifccartesianpoint`
.. [direction] :ifc:`ifcdirection`
.. [vector] :ifc:`ifcvector`
.. [line] :ifc:`ifcline`
.. [plane] :ifc:`ifcplane`
.. [axis2placement3d] :ifc:`ifcaxis2placement3d`
.. [cartesiantransformationoperator3d] :ifc:`ifccartesiantransformationoperator3d`

.. [gridplacement] :ifc:`ifcgridplacement`
.. [localplacement] :ifc:`ifclocalplacement`

.. [advancedbrep] :ifc:`ifcadvancedbrep`
.. [advancedbrepwithvoids] :ifc:`ifcadvancedbrepwithvoids`
.. [boundingbox] :ifc:`ifcboundingbox`

"""

from .geometry import IfcCartesianPoint_to_point  # noqa: F401
from .geometry import IfcDirection_to_vector  # noqa: F401
from .geometry import IfcVector_to_vector  # noqa: F401
from .geometry import IfcLine_to_line  # noqa: F401
from .geometry import IfcPlane_to_plane  # noqa: F401
from .geometry import IfcAxis2Placement3D_to_frame  # noqa: F401
from .geometry import IfcCartesianTransformationOperator3D_to_frame  # noqa: F401

from .geometricconstraint import IfcGridPlacement_to_transformation  # noqa: F401
from .geometricconstraint import IfcLocalPlacement_to_transformation  # noqa: F401
from .geometricconstraint import IfcLocalPlacement_to_frame  # noqa: F401

from .geometricmodel import IfcAdvancedBrep_to_brep  # noqa: F401
from .geometricmodel import IfcAdvancedBrepWithVoids_to_brep  # noqa: F401
from .geometricmodel import IfcBlock_to_box  # noqa: F401
from .geometricmodel import IfcBooleanClippingResult_to_brep  # noqa: F401
from .geometricmodel import IfcBoundingBox_to_box  # noqa: F401
from .geometricmodel import IfcIndexedPolyCurve_to_lines  # noqa: F401
from .geometricmodel import IfcExtrudedAreaSolid_to_brep  # noqa: F401
from .geometricmodel import IfcFacetedBrep_to_brep  # noqa: F401
from .geometricmodel import IfcFacetedBrepWithVoids_to_brep  # noqa: F401
from .geometricmodel import IfcIndexedPolygonalFaceSet_to_brep  # noqa: F401
from .geometricmodel import IfcPolygonalBoundedHalfSpace_to_brep  # noqa: F401
from .geometricmodel import IfcPolygonalFaceSet_to_brep  # noqa: F401
from .geometricmodel import IfcTessellatedFaceSet_to_brep  # noqa: F401
from .geometricmodel import IfcTriangulatedFaceSet_to_brep  # noqa: F401
from .geometricmodel import IfcShape_to_brep  # noqa: F401
from .geometricmodel import IfcShape_to_tessellatedbrep  # noqa: F401

from .shapes import frame_to_ifc_axis2_placement_3d  # noqa: F401

# from .geometricmodel import IfcBoxedHalfSpace
# from .geometricmodel import IfcCartesianPointList



from .IfcObjectDefinition import IfcObjectDefinition  # noqa: F401
from .IfcObject import IfcObject  # noqa: F401
from .IfcContext import IfcContext  # noqa: F401
from .IfcElement import IfcElement  # noqa: F401
from .IfcSpatialElement import IfcSpatialElement  # noqa: F401
from .IfcSpatialStructureElement import IfcSpatialStructureElement  # noqa: F401
from .IfcProduct import IfcProduct  # noqa: F401
from .IfcProject import IfcProject  # noqa: F401
from .IfcSite import IfcSite  # noqa: F401
from .IfcBuilding import IfcBuilding  # noqa: F401



import numpy as np
from compas.colors import Color
from compas.datastructures import Mesh
from compas_viewer.scene import ViewerSceneObject

from .tessellatedbrep import TessellatedBrep


class TessellatedBrepObject(ViewerSceneObject):
    def __init__(self, facecolors=None, **kwargs):
        super().__init__(**kwargs)

        # NOTE: it is not facecolors, it is verexcolor
        if not facecolors:
            self.facecolors = [Color(0.9, 0.9, 0.9) for _ in range(len(self.tessellatedbrep.faces) * 3)]
        else:
            facecolors = np.array(facecolors)
            self.facecolors = facecolors
            if np.mean(facecolors[:, 3]) < 1:
                # If mean alpha is less than 1, means the object has transparency
                self.opacity = 0.999  # Trigger the render order sorting of object

    @property
    def tessellatedbrep(self) -> TessellatedBrep:
        return self.item

    @property
    def bounding_box_center(self):
        if self._bounding_box_center is None:
            self._bounding_box_center = self.tessellatedbrep.vertices.mean(axis=0)
        return self._bounding_box_center

    def _read_points_data(self):
        pass

    def _read_lines_data(self):
        positions = self.tessellatedbrep.vertices.tolist()
        elements = self.tessellatedbrep.edges.tolist()
        colors = [Color(0.1, 0.1, 0.1)] * len(elements)
        return positions, colors, elements

    def _read_frontfaces_data(self):
        positions = self.tessellatedbrep.vertices[self.tessellatedbrep.faces].reshape(-1, 3).tolist()
        elements = np.arange(len(positions) * 3).reshape(-1, 3).tolist()
        colors = [Color(*color) for color in self.facecolors]
        return positions, colors, elements

    def _read_backfaces_data(self):
        positions = self.tessellatedbrep.vertices[self.tessellatedbrep.faces].reshape(-1, 3).tolist()
        elements = np.arange(len(positions) * 3).reshape(-1, 3)
        elements = elements[:, ::-1].tolist()
        colors = [Color(*color) for color in self.facecolors]
        return positions, colors, elements

    def to_mesh(self):
        return Mesh.from_vertices_and_faces(self.tessellatedbrep.vertices, self.tessellatedbrep.faces)



import os
from typing import List
from uuid import uuid4

import ifcopenshell
from compas.data import json_dump
from compas.data import json_load

from compas_ifc.entities import DEFAULT_ENTITY_TYPES
from compas_ifc.entities.building import Building
from compas_ifc.entities.buildingelements import BuildingElement
from compas_ifc.entities.buildingelements import BuildingElementProxy
from compas_ifc.entities.buildingstorey import BuildingStorey
from compas_ifc.entities.element import Element
from compas_ifc.entities.entity import Entity
from compas_ifc.entities.geographicelement import GeographicElement
from compas_ifc.entities.project import Project
from compas_ifc.entities.site import Site

from .reader import IFCReader
from .writer import IFCWriter


class Model:
    """
    Class representing an entire IFC model.

    Parameters
    ----------
    filepath : str
        The path to the IFC file.
    entity_types : dict
        A dictionary mapping IFC entity types to corresponding compas_ifc.entities classes.
        This can be used to customize the classes used to represent IFC entities.

    Attributes
    ----------
    reader : :class:`IFCReader`
        The reader used to read the content of an IFC file.
    writer : :class:`IFCWriter`
        The writer used to write the content of an IFC file.
    schema : str
        The schema of the IFC file.
    projects : List[:class:`Project`]
        The projects contained in the model.
        Typically there is exactly one.
    project : :class:`Project`
        The project of the model.
    sites : List[:class:`Site`]
        The sites contained in the model.
        In a correctly formed hierarchical model, sites are part of a project.
        If that is not the case, they are also accessible here.
    buildings : List[:class:`Building`]
        The buildings contained in the model.
        In a correctly formed hierarchical model, buildings are part of a site.
        If that is not the case, they are also accessible here.
    building_storeys : List[:class:`BuildingStorey`]
        The building storeys contained in the model.
    elements : List[:class:`BuildingElement`]
        All the building elements contained in the model.
        In a correctly formed hierarchical model, building elements are part of a building, building storey, or space.
        If that is not the case, they are also accessible here.


    """

    def __init__(self, filepath: str = None, entity_types: dict = None, use_occ=False, schema=None, load_geometries=True) -> None:
        self.reader = IFCReader(model=self, entity_types=entity_types, use_occ=use_occ)
        self.writer = IFCWriter(model=self)
        self._new_entities = set()
        self._projects = None
        self._sites = None
        self._buildings = None
        self._building_storeys = None
        self._elements = None
        self._building_elements = None
        self._geographic_elements = None
        self._schema = None
        if schema:
            self._schema = ifcopenshell.ifcopenshell_wrapper.schema_by_name(schema)

        if filepath:
            self.open(filepath, load_geometries=load_geometries)

    def open(self, filepath: str, load_geometries=True) -> None:
        self.reader.open(filepath, load_geometries=load_geometries)

    def save(self, filepath: str) -> None:
        self.writer.save(filepath)

    def export(self, entities, filepath: str) -> None:
        self.writer.export(entities, filepath)

    def get_all_entities(self) -> List[Entity]:
        """Get all entities in the model."""
        return self.reader.get_all_entities() + list(self._new_entities)

    def get_entities_by_type(self, ifc_type: str, include_subtypes: bool = True, sort_by_name=True) -> List[Entity]:
        """Get all entities of a specific ifc type in the model. If include_subtypes is True, also return entities of subtypes of the given type."""
        entities = self.reader.get_entities_by_type(ifc_type, include_subtypes)
        for entity in self._new_entities:
            if entity.is_a(ifc_type):
                entities.append(entity)

        if sort_by_name:
            entities.sort(key=lambda x: x.name if x.name else "")
        return entities

    def get_entity_by_global_id(self, global_id) -> Entity:
        """Get an entity by its global id."""
        return self.reader.get_entity_by_global_id(global_id)

    def get_entity_by_id(self, name) -> Entity:
        """Get an entity by its id in the IFC file."""
        return self.reader.get_entity_by_id(name)

    def get_entities_by_name(self, entity_name) -> List[Entity]:
        """Get all entities with a specific name."""
        return self.reader.get_entities_by_name(entity_name)

    def print_spatial_hierarchy(self, max_level: int = 4) -> None:
        """Print the spatial hierarchy of the model."""
        self.project.print_spatial_hierarchy(max_level)

    def print_summary(self):
        """Print a summary of the model."""

        print("=" * 80)
        print("File: {}".format(self.reader.filepath))
        # print("Size: {} MB".format(self.reader.file_size()))
        print("Project: {}".format(self.project.name))
        print("Description: {}".format(self.project.attributes.get("Description", "")))
        print("Number of sites: {}".format(len(self.sites)))
        print("Number of buildings: {}".format(len(self.buildings)))
        print("Number of building elements: {}".format(len(self.building_elements)))
        print("=" * 80)

    @property
    def schema(self) -> str:
        return self._schema or self.reader._schema

    @property
    def project(self) -> Project:
        if self.projects:
            return self.projects[0]

    @property
    def projects(self) -> List[Project]:
        if self._projects is None:
            self._projects = self.get_entities_by_type("IfcProject")
        return self._projects

    @property
    def sites(self) -> List[Site]:
        if self._sites is None:
            self._sites = self.get_entities_by_type("IfcSite")
        return self._sites

    @property
    def buildings(self) -> List[Building]:
        if self._buildings is None:
            self._buildings = self.get_entities_by_type("IfcBuilding")
        return self._buildings

    @property
    def building_storeys(self) -> List[BuildingStorey]:
        if self._building_storeys is None:
            self._building_storeys = self.get_entities_by_type("IfcBuildingStorey")
        return self._building_storeys

    @property
    def elements(self) -> List[Element]:
        if self._elements is None:
            self._elements = self.get_entities_by_type("IfcElement")
        return self._elements

    @property
    def building_elements(self) -> List[BuildingElement]:
        if self._building_elements is None:
            self._building_elements = self.get_entities_by_type("IfcBuildingElement")
        return self._building_elements

    @property
    def geographic_elements(self) -> List[GeographicElement]:
        if self._geographic_elements is None:
            self._geographic_elements = self.get_entities_by_type("IfcGeographicElement")
        return self._geographic_elements

    def create(self, cls=None, parent=None, geometry=None, frame=None, psets=None, **kwargs):
        """Create an entity and add it to the model.

        Parameters
        ----------
        cls : :class:`compas_ifc.entities.Entity`
            The type of entity to create.
        attributes : dict
            The attributes of the entity.
        parent : :class:`compas_ifc.entities.Entity`
            The parent entity of the entity.

        Returns
        -------
        :class:`compas_ifc.entities.Entity`
            The created entity.
        """

        def get_type(type_name):
            types = DEFAULT_ENTITY_TYPES.copy()
            if self.reader.entity_types:
                types.update(self.reader.entity_types)

            for key, value in types.items():
                if key == type_name:
                    return value

            return BuildingElementProxy

        if not isinstance(cls, type):
            cls = get_type(cls)

        entity = cls(None, self)
        entity.set_attributes(kwargs)

        if frame:
            if not hasattr(entity, "frame"):
                raise TypeError(f"{cls} cannot be assigned a frame.")
            entity.frame = frame

        if geometry:
            if not hasattr(entity, "body"):
                raise TypeError(f"{cls} cannot be assigned a geometry.")
            entity.body = geometry

        if parent:
            if not hasattr(entity, "parent"):
                raise TypeError(f"{cls} cannot be assigned a parent.")
            entity.parent = parent

        if psets:
            entity.psets = psets

        self._new_entities.add(entity)

        if cls == Project:
            self._projects = [entity]

        return entity

    def insert(self, geometry, parent=None, name=None, description=None, cls=None, frame=None) -> BuildingElementProxy:
        """Insert a geometry into the model. The geometry will be wrapped in a building element proxy.

        Parameters
        ----------
        geometry : :class:`compas.geometry.Geometry` or :class:`compas.datastructures.Mesh`
            The geometry to be inserted.
        parent : :class:`compas_ifc.entities.Entity`
            The parent entity of the geometry.
        name : str
            The name of the element.
        description : str
            The description of the element.

        Returns
        -------
        :class:`compas_ifc.entities.buildingelements.BuildingElementProxy`
            The building element proxy wrapping the geometry.
        """
        if cls is None:
            cls = BuildingElementProxy
        element = cls(None, self)
        element.body = geometry
        element.parent = parent
        element.frame = frame
        element["Name"] = name
        element["Description"] = description
        self._new_entities.add(element)
        return element

    def export_session(self, path, export_geometries=True):
        try:
            import shutil

            from compas.datastructures import Mesh
            from compas.geometry import Box
            from compas.geometry import Shape
            from compas.geometry import Sphere
            from compas_occ.brep import Brep

        except ImportError:
            raise ImportError("The export_session method requires compas_occ to be installed.")

        if os.path.exists(path):
            if input(f"Directory {path} already exists. Do you want to overwrite it? (Y/n): ") != "n":
                shutil.rmtree(path)
            else:
                return

        os.makedirs(path, exist_ok=True)
        os.makedirs(os.path.join(path, "geometries"), exist_ok=True)

        geometry_map = {}

        def export_entity(entity: Entity, geometry_map=geometry_map):
            data = {}
            data["type"] = entity.ifc_type
            data["name"] = entity.name
            data["description"] = entity["Description"]

            if entity.psets:
                data["psets"] = entity.psets

            if entity.children:
                data["children"] = []

                for child in entity.children:
                    data["children"].append(export_entity(child))

            if export_geometries and hasattr(entity, "body_with_opening") and entity.body_with_opening:
                if geometry_map.get(id(entity.body_with_opening)):
                    # TODO: check when reading from existing IFC files.
                    data["geometry"] = geometry_map[id(entity.body_with_opening)]
                else:
                    if isinstance(entity.body_with_opening, Box):
                        brep = Brep.from_box(entity.body_with_opening)
                    elif isinstance(entity.body_with_opening, Sphere):
                        brep = Brep.from_sphere(entity.body_with_opening)
                    elif isinstance(entity.body_with_opening, Mesh):
                        brep = Brep.from_mesh(entity.body_with_opening)
                    elif isinstance(entity.body_with_opening, Shape):
                        brep = Brep.from_mesh(Mesh.from_shape(entity.body_with_opening))
                    elif isinstance(entity.body_with_opening, Brep):
                        brep = entity.body_with_opening
                    else:
                        raise ValueError(f"Geometry type {type(entity.body_with_opening)} not supported.")

                    file_name = f"{uuid4()}.stp"
                    brep.to_step(os.path.join(path, "geometries", file_name))
                    data["geometry"] = file_name
                    geometry_map[id(entity.body_with_opening)] = file_name

            if entity.frame:
                data["frame"] = entity.frame

            return data

        data = export_entity(self.project, geometry_map=geometry_map)
        json_dump(data, os.path.join(path, "session.json"), pretty=False)

    def load_session(self, path):
        try:
            from compas_occ.brep import Brep

            from compas_ifc.entities import DEFAULT_ENTITY_TYPES

        except ImportError:
            raise ImportError("The load_session method requires compas_occ to be installed.")

        session = json_load(os.path.join(path, "session.json"))
        geometry_map = {}

        def get_type(type_name):
            types = DEFAULT_ENTITY_TYPES.copy()
            if self.reader.entity_types:
                types.update(self.reader.entity_types)

            for key, value in types.items():
                if key == type_name:
                    return value

            return BuildingElementProxy

        def load_entity(data, parent=None, geometry_map=geometry_map):
            ifc_type = get_type(data["type"])
            entity = self.create(ifc_type, parent=parent, Name=data["name"], Description=data.get("description", ""))

            if "children" in data:
                for child_data in data["children"]:
                    load_entity(child_data, entity)

            if "geometry" in data:
                if geometry_map.get(data["geometry"]):
                    entity.body = geometry_map.get(data["geometry"])
                else:
                    entity.body = Brep.from_step(os.path.join(path, "geometries", data["geometry"]))
                    geometry_map[data["geometry"]] = entity.body

            if "frame" in data:
                print(entity)
                entity.frame = data["frame"]

            if "psets" in data:
                entity.psets = data["psets"]

        load_entity(session, parent=None, geometry_map=geometry_map)

    @classmethod
    def template(cls, schema="IFC4", building_count=1, storey_count=1):
        model = cls(schema=schema)
        project = model.create(Project, Name="Default Project")
        site = model.create(Site, parent=project, Name="Default Site")
        for i in range(building_count):
            building = model.create(Building, parent=site, Name=f"Default Building {i+1}")
            for j in range(storey_count):
                model.create(BuildingStorey, parent=building, Name=f"Default Storey {j+1}")
        return model

    def show(self, entity=None):
        try:
            from compas_viewer import Viewer

            # from compas.datastructures import Tree
            # from compas.datastructures import TreeNode
            from compas_viewer.config import Config

            # from compas_viewer.components.treeform import Treeform
            # from compas_viewer.components import Sceneform
        except ImportError:
            raise ImportError("The show method requires compas_viewer to be installed.")

        config = Config()
        # config.ui.sidebar.sceneform = False
        # config.ui.sidedock.show = False

        viewer = Viewer(config=config)

        entity_map = {}

        def parse_entity(entity, parent=None):
            obj = None
            if getattr(entity, "geometry", None):
                if not entity.is_a("IfcSpace"):
                    obj = viewer.scene.add(entity.geometry, name=entity.name, parent=parent, **entity.style)
            if entity.children:
                obj = viewer.scene.add([], name=entity.name, parent=parent)
                for child in entity.children:
                    parse_entity(child, parent=obj)

            if obj:
                entity_map[id(obj)] = entity

        parse_entity(entity or self.project)

        # propertyform = Treeform(Tree(), {"Name": (lambda o: o.name), "value": (lambda o: o.attributes.get("value", ""))})

        # def callback(obj):

        #     entity = entity_map[id(obj)]
        #     tree = Tree()
        #     root = TreeNode("ROOT")
        #     tree.add(root)
        #     attribute_node = TreeNode("Attributes")
        #     root.add(attribute_node)
        #     for name, value in entity.attributes.items():
        #         attribute_node.add(TreeNode(name, value=value))

        #     properties_node = TreeNode("Properties")
        #     root.add(properties_node)

        #     for pset, properties in entity.psets.items():
        #         pset_node = TreeNode(pset)
        #         properties_node.add(pset_node)
        #         for name, value in properties.items():
        #             pset_node.add(TreeNode(name, value=value))

        #     propertyform.tree = tree
        #     propertyform.update()

        # def get_name(obj):
        #     entity = entity_map.get(id(obj))
        #     if entity:
        #         return f"[{entity.ifc_type}] {entity.name}"
        #     return ""

        # sceneform = Sceneform(viewer.scene, {"Name": get_name}, callback=callback)

        # viewer.ui.sidebar.widget.addWidget(sceneform)
        # viewer.ui.sidebar.widget.addWidget(propertyform)

        viewer.show()



import ifcopenshell

from .entity import Entity


class Root(Entity):
    """Base class for all IFC entities that are derived from IfcRoot.

    Attributes
    ----------
    global_id : str
        The global id of the entity.
    name : str
        The name of the entity.
    """

    def __init__(self, entity, model) -> None:
        super().__init__(entity, model)
        if not entity:
            self["GlobalId"] = ifcopenshell.guid.new()
            self["Name"] = None

    def __repr__(self):
        return "<{} Name: {}, GlobalId: {}>".format(self.ifc_type, self.name, self.global_id)

    @property
    def global_id(self):
        return self["GlobalId"]

    @property
    def name(self):
        return self["Name"]



class IfcObjectDefinition:
    def test(self):
        print("test")


class IfcBuilding:
    def test2(self):
        print("test2")


EXTENSIONS = {
    "IfcObjectDefinition": IfcObjectDefinition,
    "IfcBuilding": IfcBuilding,
}



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from compas_ifc.entities.generated.IFC4 import IfcProject
else:
    IfcProject = object


class IfcProject(IfcProject):
    @property
    def sites(self):
        return self.children_by_type("IfcSite", recursive=True)

    @property
    def buildings(self):
        return self.children_by_type("IfcBuilding", recursive=True)

    @property
    def building_elements(self):
        return self.children_by_type("IfcBuildingElement", recursive=True)

    @property
    def geographic_elements(self):
        return self.children_by_type("IfcGeographicElement", recursive=True)

    @property
    def contexts(self):
        from compas.geometry import Vector

        from compas_ifc.resources import IfcAxis2Placement3D_to_frame

        contexts = []

        for context in self.RepresentationContexts:
            north = Vector(*context.TrueNorth.DirectionRatios) if context.TrueNorth else None
            wcs = IfcAxis2Placement3D_to_frame(context.WorldCoordinateSystem)
            contexts.append(
                {
                    "identifier": context.ContextIdentifier,
                    "type": context.ContextType,
                    "precision": context.Precision,
                    "dimension": context.CoordinateSpaceDimension,
                    "north": north,
                    "wcs": wcs,
                }
            )
        return contexts

    @property
    def units(self):
        units = []
        units_in_context = self.UnitsInContext or self.file.get_entities_by_type("IfcUnitAssignment")[0]
        for unit in units_in_context.Units:
            if unit.is_a("IfcSIUnit") or unit.is_a("IfcConversionBasedUnit"):
                units.append(unit)
        return units

    @property
    def length_unit(self):
        for unit in self.units:
            if unit["UnitType"] == "LENGTHUNIT":
                return unit

    @property
    def length_scale(self):
        unit = self.length_unit
        if unit:
            if unit.Name == "METRE" and not unit.Prefix:
                return 1.0
            if unit.Name == "METRE" and unit.Prefix == "CENTI":
                return 1e-2
            if unit.Name == "METRE" and unit.Prefix == "MILLI":
                return 1e-3
        return 1.0

    @property
    def frame(self):
        for context in self.contexts:
            if context["type"] == "Model":
                return context["wcs"]

    @property
    def north(self):
        for context in self.contexts:
            if context["type"] == "Model":
                return context["north"]



import importlib
from typing import TYPE_CHECKING

from compas.data import Data
from compas.datastructures import Tree
from compas.datastructures import TreeNode
from ifcopenshell import entity_instance

if TYPE_CHECKING:
    from compas_ifc.file import IFCFile
    from compas_ifc.model import Model


class TypeDefinition:
    def __init__(self, entity: entity_instance = None, file=None):
        self.file = file
        self.entity = entity

    @property
    def value(self):
        return self.entity.wrappedValue

    def __repr__(self):
        return "<{} {}>".format(self.entity.is_a(), self.value)


class Base(Data):
    """Base class for all IFC entities.

    Attributes
    ----------
    entity : entity_instance
        The IFC entity instance.
    file : Ifcfile
    """

    def __new__(cls, entity: entity_instance, file: "IFCFile" = None):
        if file is None:
            schema = "IFC4"
        else:
            schema = file._schema.name()

        try:
            classes = importlib.import_module(f"compas_ifc.entities.generated.{schema}")
        except ImportError:
            raise ImportError(f"compas_ifc classes for schema {schema} was not generated. Run `python -m compas_ifc.entities.generator`")

        cls_name = entity.is_a()
        ifc_cls = getattr(classes, cls_name, None)
        if ifc_cls:
            return super(Base, cls).__new__(ifc_cls)
        elif hasattr(entity, "wrappedValue"):
            return TypeDefinition(entity, file)

    def __init__(self, entity: entity_instance = None, file=None):
        super().__init__()
        self.file = file
        self.entity = entity

    def __repr__(self):
        return "<#{} {}>".format(self.entity.id(), self.__class__.__name__)

    def __getitem__(self, key):
        return getattr(self, key)

    def __iter__(self):
        return iter(self.all_attribute_names())

    def _get_attribute(self, name=None, entity: entity_instance = None):
        if name is not None:
            attr = getattr(self.entity, name)
        else:
            attr = entity
        if isinstance(attr, entity_instance):
            # NOTE: Double check.
            # if hasattr(attr, "wrappedValue"):
            #     return attr.wrappedValue
            return self.file.from_entity(attr)
        if isinstance(attr, (list, tuple)):
            return [self._get_attribute(entity=item) for item in attr]
        else:
            return attr

    def _set_attribute(self, name, value):
        # TODO: re-enable strong type checking
        # cls = self.__class__
        # getter_type_hints = get_type_hints(getattr(cls, name).fget)
        # value_type = getter_type_hints["return"]

        # if hasattr(value_type, "__origin__"):
        #     # deal parameterized generic types
        #     origin = value_type.__origin__
        #     if origin == list:
        #         if not isinstance(value, (list, tuple)):
        #             raise TypeError(f"Expected {value_type}, got {type(value)} for {cls.__name__}.{name}")
        #         value_type = getter_type_hints["return"].__args__[0]
        #         if not all(isinstance(item, value_type) for item in value):
        #             raise TypeError(f"Expected {value_type}, got {type(value)} for {cls.__name__}.{name}")
        #     else:
        #         raise NotImplementedError(f"Unsupported generic type {origin}")

        # elif not isinstance(value, value_type):
        #     raise TypeError(f"Expected {value_type}, got {type(value)} for {cls.__name__}.{name}")

        def prepare_value(value):
            if isinstance(value, Base):
                return value.entity
            elif isinstance(value, TypeDefinition):
                return value.entity
            else:
                return value

        if isinstance(value, (list, tuple)):
            value = [prepare_value(v) for v in value]
        else:
            value = prepare_value(value)

        if getattr(self.entity, name) != value:
            try:
                setattr(self.entity, name, value)
            except Exception as e:
                print(f"Error setting {name} of {self} to {value}")
                raise e

    def _get_inverse_attribute(self, name):
        return [self.file.from_entity(attr) for attr in getattr(self.entity, name)]

    @property
    def model(self) -> "Model":
        return self.file.model  # TODO: rather convoluted.

    @property
    def schema(self):
        return self.file._schema.name()

    def id(self):
        return self.entity.id()

    def is_a(self, type_name=None):
        if type_name:
            return self.entity.is_a(type_name)
        else:
            return self.entity.is_a()

    def all_attribute_names(self):
        # all_attributes = []
        # def get_attr_names(cls):
        #     if cls.__name__ == "Base":
        #         return
        #     all_attributes.extend(cls.attributes)
        #     for base in cls.__bases__:
        #         get_attr_names(base)
        # get_attr_names(self.__class__)
        # return all_attributes
        info = self.entity.get_info(include_identifier=False)
        del info["type"]
        return list(info.keys())

    @property
    def attributes(self):
        return self.to_dict()

    def to_dict(self, recursive=False, ignore_fields=[], include_fields=[], convert_type_definitions=False):
        def iter_list(values):
            _values = []
            for value in values:
                if isinstance(value, Base):
                    value = value.to_dict(recursive=recursive, ignore_fields=ignore_fields, include_fields=include_fields)
                elif convert_type_definitions and isinstance(value, TypeDefinition):
                    value = value.value
                elif isinstance(value, (list, tuple)):
                    value = iter_list(value)
                _values.append(value)
            return _values

        data = {}
        for key in self:
            if key in ignore_fields:
                continue

            if include_fields and key not in include_fields:
                continue

            value = getattr(self, key)

            if recursive and isinstance(value, Base):
                value = value.to_dict(recursive=recursive, ignore_fields=ignore_fields, include_fields=include_fields)
            elif recursive and isinstance(value, (list, tuple)):
                value = iter_list(value)
            elif convert_type_definitions and isinstance(value, TypeDefinition):
                value = value.value

            data[key] = value

        return data

    def print_attributes(self, max_depth=2):
        attr_tree = Tree()
        root = EntityNode(name=f"ROOT: {self} [{self.__class__.__name__}]")
        attr_tree.add(root)

        def add_entity(entity, node, depth):
            if depth < max_depth:
                if isinstance(entity, list):
                    entity = dict(enumerate(entity))
                for key in entity:
                    sub_entity = entity[key]
                    sub_node = EntityNode(name=f"{key}: {sub_entity} [{sub_entity.__class__.__name__}]")
                    node.add(sub_node)
                    if isinstance(sub_entity, (Base, list)):
                        add_entity(sub_entity, sub_node, depth + 1)

        add_entity(self, root, 0)

        print("=" * 80 + "\n" + f"Attributes of {self}\n" + "=" * 80)
        print(attr_tree.get_hierarchy_string(max_depth=max_depth))
        print("")

    def attribute_info(self):
        raise NotImplementedError

    def print_spatial_hierarchy(self, max_depth=5):
        IfcObjectDefinition = getattr(importlib.import_module(f"compas_ifc.entities.generated.{self.schema}"), "IfcObjectDefinition")

        if not isinstance(self, IfcObjectDefinition):
            raise TypeError("Only IfcObjectDefinition has spatial hierarchy")

        top = self
        while top.parent:
            top = top.parent

        spatial_tree = Tree()

        def add_entity(entity, parent_node):
            for child_entity in entity.children:
                name = f"{child_entity}"
                if child_entity == self:
                    name = "**" + name + "**"
                node = EntityNode(name=name)
                parent_node.add(node)
                add_entity(child_entity, node)

        root_node = EntityNode(name=f"{top}")
        spatial_tree.add(root_node)
        add_entity(top, root_node)

        print("=" * 80 + "\n" + f"Spatial hierarchy of {self}\n" + "=" * 80)
        print(spatial_tree.get_hierarchy_string(max_depth=max_depth))
        print("")

    def print_properties(self):
        IfcObject = getattr(importlib.import_module(f"compas_ifc.entities.generated.{self.schema}"), "IfcObject")

        if not isinstance(self, IfcObject):
            raise TypeError("Only IfcObject has properties")

        def add_property(item, parent_node):
            if isinstance(item, dict):
                for key, value in item.items():
                    if isinstance(value, dict):
                        node = EntityNode(name=f"{key}")
                        parent_node.add(node)
                        add_property(value, node)
                    else:
                        node = EntityNode(name=f"{key}: {value}")
                        parent_node.add(node)

        tree = Tree()
        root = EntityNode(name=f"{self}")
        tree.add(root)
        add_property(self.properties, root)

        print("=" * 80 + "\n" + f"Properties of {self}\n" + "=" * 80)
        tree.print_hierarchy()
        print("")

    def show(self):
        self.model.show(self)


class EntityNode(TreeNode):
    def __repr__(self):
        return self.name



import ifcopenshell
from compas.geometry import Box
from compas.geometry import Cone
from compas.geometry import Cylinder
from compas.geometry import Sphere


def frame_to_ifc_axis2_placement_3d(file, frame):
    return create_IfcAxis2Placement3D(file, point=frame.point, dir1=frame.zaxis, dir2=frame.xaxis)


def create_IfcAxis2Placement3D(file, point=None, dir1=None, dir2=None):
    """
    Create an IFC Axis2Placement3D from a point, a direction and a second direction.
    """
    point = file.createIfcCartesianPoint(point or (0.0, 0.0, 0.0))
    dir1 = file.createIfcDirection(dir1 or (0.0, 0.0, 1.0))
    dir2 = file.createIfcDirection(dir2 or (1.0, 0.0, 0.0))
    axis2placement = file.createIfcAxis2Placement3D(point, dir1, dir2)
    return axis2placement


def create_IfcShapeRepresentation(file: ifcopenshell.file, item: ifcopenshell.entity_instance, context: ifcopenshell.entity_instance) -> ifcopenshell.entity_instance:
    """
    Create an IFC Shape Representation from an IFC item and a context.
    """
    return file.create_entity(
        "IfcShapeRepresentation",
        ContextOfItems=context,
        RepresentationIdentifier="Body",
        RepresentationType="SolidModel",
        Items=[item],
    )


def box_to_IfcBlock(file: ifcopenshell.file, box: Box) -> ifcopenshell.entity_instance:
    """
    Convert a COMPAS box to an IFC Block.
    """
    pt = box.frame.point.copy()
    pt -= [box.xsize / 2, box.ysize / 2, box.zsize / 2]
    return file.create_entity(
        "IfcBlock",
        Position=create_IfcAxis2Placement3D(file, pt, box.frame.zaxis, box.frame.xaxis),
        XLength=box.xsize,
        YLength=box.ysize,
        ZLength=box.zsize,
    )


def sphere_to_IfcSphere(file: ifcopenshell.file, sphere: Sphere) -> ifcopenshell.entity_instance:
    """
    Convert a COMPAS sphere to an IFC Sphere.
    """
    return file.create_entity(
        "IfcSphere",
        Position=create_IfcAxis2Placement3D(file, sphere.base),
        Radius=sphere.radius,
    )


def cone_to_IfcRightCircularCone(file: ifcopenshell.file, cone: Cone) -> ifcopenshell.entity_instance:
    """
    Convert a COMPAS cone to an IFC Cone.
    """
    plane = cone.circle.plane
    return file.create_entity(
        "IfcRightCircularCone",
        Position=create_IfcAxis2Placement3D(file, plane.point, plane.normal),
        Height=cone.height,
        BottomRadius=cone.circle.radius,
    )


def cylinder_to_IfcRightCircularCylinder(file: ifcopenshell.file, cylinder: Cylinder) -> ifcopenshell.entity_instance:
    """
    Convert a COMPAS cylinder to an IFC Cylinder.
    """
    plane = cylinder.circle.plane
    return file.create_entity(
        "IfcRightCircularCylinder",
        Position=create_IfcAxis2Placement3D(file, plane.point, plane.normal),
        Height=cylinder.height,
        Radius=cylinder.circle.radius,
    )


def occ_cylinder_to_ifc_cylindrical_surface(file, occ_cylinder):
    location = occ_cylinder.Location().Coord()
    xdir = occ_cylinder.XAxis().Direction().Coord()
    zdir = occ_cylinder.Axis().Direction().Coord()
    IfcAxis2Placement3D = create_IfcAxis2Placement3D(file, location, zdir, xdir)
    return file.create_entity("IfcCylindricalSurface", IfcAxis2Placement3D, occ_cylinder.Radius())



from typing import TYPE_CHECKING

from ifcopenshell.util.element import get_material

if TYPE_CHECKING:
    from compas_ifc.entities.generated.IFC4 import IfcObjectDefinition
else:
    IfcObjectDefinition = object


class IfcObjectDefinition(IfcObjectDefinition):
    def __repr__(self):
        return '<#{} {} "{}">'.format(self.entity.id(), self.__class__.__name__, self.Name)

    @property
    def parent(self):
        relations = self.Decomposes()
        if relations:
            return relations[0].RelatingObject
        else:
            return None

    @property
    def children(self):
        return sum([relation.RelatedObjects for relation in self.IsDecomposedBy()], [])

    @property
    def descendants(self):
        descendants = []
        for child in self.children:
            descendants.append(child)
            descendants.extend(child.descendants)
        return descendants

    @property
    def material(self):
        material = get_material(self.entity)
        if material:
            return self.file.from_entity(material)
        else:
            return None

    def children_by_type(self, type_name, recursive=False):
        if not recursive:
            return [child for child in self.children if child.is_a(type_name)]
        else:
            return [child for child in self.descendants if child.is_a(type_name)]



"""
********************************************************************************
"attributes"
********************************************************************************

.. currentmodule:: compas_ifc.attributes

Functions
=========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    is_attribute_aggregation
    is_attribute_enumeration
    attribute_enumeration_items
    declared_object_attributes
    declared_context_attributes
    declared_product_attributes
    declared_building_element_attributes
    declared_spatial_element_attributes

"""


def is_attribute_aggregation(attribute):
    """
    Verify that an attribute is an aggregation type.

    Parameters
    ----------
    attribute

    Returns
    -------
    bool

    """
    return attribute.type_of_attribute().as_aggregation_type()


def is_attribute_enumeration(attribute):
    """
    Verify that an attribute is an enumeration type.

    Parameters
    ----------
    attribute

    Returns
    -------
    bool

    """
    if attribute.type_of_attribute().as_named_type():
        return attribute.type_of_attribute().declared_type().as_enumeration_type()
    return False


def attribute_enumeration_items(attribute):
    """
    Get the value options of an enumeration type.

    Parameters
    ----------
    attribute

    Returns
    -------
    tuple

    """
    enumeration = attribute.type_of_attribute().declared_type().as_enumeration_type()
    return enumeration.enumeration_items()


def declared_object_attributes(schema):
    """
    Get all declared attributes of an IFC object.

    Parameters
    ----------
    schema

    Returns
    -------
    tuple

    """
    declaration = schema.declaration_by_name("IfcObject")
    return declaration.all_attributes()


def declared_context_attributes(schema):
    declaration = schema.declaration_by_name("IfcContext")
    return declaration.all_attributes()


def declared_product_attributes(schema):
    declaration = schema.declaration_by_name("IfcProduct")
    return declaration.all_attributes()


def declared_building_element_attributes(schema, exclude_product=True):
    declaration = schema.declaration_by_name("IfcBuildingElement")
    if not exclude_product:
        return declaration.all_attributes()

    attributes = []
    product = [attr.name() for attr in declared_product_attributes(schema)]
    for attribute in declaration.all_attributes():
        if attribute.name() not in product:
            attributes.append(attribute)
    return attributes


def declared_spatial_element_attributes(schema, exclude_product=True):
    declaration = schema.declaration_by_name("IfcSpatialElement")
    if not exclude_product:
        return declaration.all_attributes()

    attributes = []
    product = declared_product_attributes(schema)
    for attribute in declaration.all_attributes():
        if attribute not in product:
            attributes.append(attribute)
    return attributes



from typing import List

from compas_ifc.entities.buildingelements import Beam
from compas_ifc.entities.buildingelements import BuildingElementProxy
from compas_ifc.entities.buildingelements import Column
from compas_ifc.entities.buildingelements import Door
from compas_ifc.entities.buildingelements import Roof
from compas_ifc.entities.buildingelements import Slab
from compas_ifc.entities.buildingelements import Stair
from compas_ifc.entities.buildingelements import Wall
from compas_ifc.entities.buildingelements import Window
from compas_ifc.entities.space import Space
from compas_ifc.entities.spatialelement import SpatialElement


class BuildingStorey(SpatialElement):
    """
    Class representing an IFC building storey.

    Attributes
    ----------
    beams : List[:class:`Beam`]
        The beams contained in this building storey.
    columns : List[:class:`Column`]
        The beams contained in this building storey.
    doors : List[:class:`Column`]
        The beams contained in this building storey.
    roofs : List[:class:`Roof`]
        The roofs contained in this building storey.
    slabs : List[:class:`Slab`]
        The slabs contained in this building storey.
    stairs : List[:class:`Stairs`]
        The stairs contained in this building storey.
    walls : List[:class:`Walls`]
        The walls contained in this building storey.
    windows : List[:class:`Windows`]
        The windows contained in this building storey.
    spaces : List[:class:`Space`]
        The spaces contained in this building storey.
    building_elements_proxies : List[:class:`BuildingElementProxy`]
        The building element proxies contained in this building storey.

    """

    @property
    def spaces(self) -> List[Space]:
        return [entity for entity in self.traverse() if entity.is_a("IfcSpace")]

    @property
    def beams(self) -> List[Beam]:
        return [entity for entity in self.traverse() if entity.is_a("IfcBeam")]

    @property
    def columns(self) -> List[Column]:
        return [entity for entity in self.traverse() if entity.is_a("IfcColumn")]

    @property
    def doors(self) -> List[Door]:
        return [entity for entity in self.traverse() if entity.is_a("IfcDoor")]

    @property
    def roofs(self) -> List[Roof]:
        return [entity for entity in self.traverse() if entity.is_a("IfcRoof")]

    @property
    def slabs(self) -> List[Slab]:
        return [entity for entity in self.traverse() if entity.is_a("IfcSlab")]

    @property
    def stairs(self) -> List[Stair]:
        return [entity for entity in self.traverse() if entity.is_a("IfcStair")]

    @property
    def walls(self) -> List[Wall]:
        return [entity for entity in self.traverse() if entity.is_a("IfcWall")]

    @property
    def windows(self) -> List[Window]:
        return [entity for entity in self.traverse() if entity.is_a("IfcWindow")]

    @property
    def building_elements_proxies(self) -> List[BuildingElementProxy]:
        return [entity for entity in self.traverse() if entity.is_a("IfcBuildingElementProxy")]



from .tessellatedbrep import TessellatedBrep

try:
    from .tessellatedbrepobject import TessellatedBrepObject
except ImportError:
    pass

from compas.plugins import plugin
from compas.scene import register


@plugin(category="factories", requires=["compas_viewer"], trylast=True)
def register_scene_objects():
    register(TessellatedBrep, TessellatedBrepObject, context="Viewer")

    try:
        from compas_occ.brep import OCCBrep
        from .ifcbrepobject import IFCBrepObject

        register(OCCBrep, IFCBrepObject, context="Viewer")

    except ImportError:
        pass


__all__ = ["TessellatedBrep", "TessellatedBrepObject"]



from compas.datastructures import Mesh
from compas.geometry import Box
from compas.geometry import Cone
from compas.geometry import Cylinder
from compas.geometry import Sphere
from compas.tolerance import TOL
from ifcopenshell.api import run

from .brep import brep_to_ifc_advanced_brep
from .mesh import mesh_to_IfcPolygonalFaceSet
from .mesh import mesh_to_IfcShellBasedSurfaceModel
from .shapes import box_to_IfcBlock
from .shapes import cone_to_IfcRightCircularCone
from .shapes import cylinder_to_IfcRightCircularCylinder
from .shapes import sphere_to_IfcSphere


def write_body_representation(file, body, ifc_entity, context):
    try:
        from compas_occ.brep import OCCBrep
    except ImportError:
        OCCBrep = None

    def _body_to_shape(body):
        if isinstance(body, Box):
            shape = box_to_IfcBlock(file, body)
        elif isinstance(body, Sphere):
            shape = sphere_to_IfcSphere(file, body)
        elif isinstance(body, Cone):
            shape = cone_to_IfcRightCircularCone(file, body)
        elif isinstance(body, Cylinder):
            shape = cylinder_to_IfcRightCircularCylinder(file, body)
        elif isinstance(body, Mesh):
            if file.schema == "IFC4" or file.schema == "IFC4x3":
                shape = mesh_to_IfcPolygonalFaceSet(file, body)
            else:
                shape = mesh_to_IfcShellBasedSurfaceModel(file, body)
        elif OCCBrep and isinstance(body, OCCBrep):
            if file.schema == "IFC2X3":
                print("IFC2X3 does not support advanced brep, converting representation of {} to mesh.".format(ifc_entity))
                shape = mesh_to_IfcShellBasedSurfaceModel(file, body.to_viewmesh(linear_deflection=TOL.lineardeflection)[0])
            else:
                shape = brep_to_ifc_advanced_brep(file, body)
        else:
            raise Exception("Unsupported body type.")
        return shape

    if isinstance(body, list):
        shape = []
        for b in body:
            s = _body_to_shape(b)
            if not isinstance(s, list):
                shape.append(s)
            else:
                shape.extend(s)
    else:
        shape = _body_to_shape(body)
        if not isinstance(shape, list):
            shape = [shape]

    RepresentationType = "SolidModel"

    if shape and shape[0].is_a("IfcShellBasedSurfaceModel"):
        RepresentationType = "SurfaceModel"

    representation = file.create_entity(
        "IfcShapeRepresentation",
        ContextOfItems=context,
        RepresentationIdentifier="Body",
        RepresentationType=RepresentationType,
        Items=shape,
    )

    run("geometry.assign_representation", file, product=ifc_entity, representation=representation)

    return representation



import multiprocessing
import os
import time
from typing import List

import ifcopenshell
import numpy as np
from compas.geometry import Transformation

from compas_ifc.entities.entity import Entity


class IFCReader(object):
    """A class for reading IFC files.

    Parameters
    ----------
    model : :class:`compas_ifc.model.Model`
        The model to which the reader belongs.
    entity_types : dict
        A dictionary mapping IFC entity types to corresponding compas_ifc.entities classes.
        This can be used to customize the classes used to represent IFC entities.

    Attributes
    ----------
    model : :class:`compas_ifc.model.Model`
        The model to which the reader belongs.
    entity_types : dict
        A dictionary mapping IFC entity types to corresponding compas_ifc.entities classes.
    projects : List[:class:`Project`]
        The projects contained in the model.
        Typically there is exactly one.
    project : :class:`Project`
        The project of the model.
    sites : List[:class:`Site`]
        The sites contained in the model.
        In a correctly formed hierarchical model, sites are part of a project.
        If that is not the case, they are also accessible here.
    buildings : List[:class:`Building`]
        The buildings contained in the model.
        In a correctly formed hierarchical model, buildings are part of a site.
        If that is not the case, they are also accessible here.
    building_storeys : List[:class:`BuildingStorey`]
        The building storeys contained in the model.
    elements : List[:class:`BuildingElement`]
        All the building elements contained in the model.
        In a correctly formed hierarchical model, building elements are part of a building, building storey, or space.
        If that is not the case, they are also accessible here.

    """

    def __init__(self, model, entity_types: dict = None, use_occ=True):
        self.filepath = None
        self.model = model
        self.entity_types = entity_types
        self.use_occ = use_occ
        self._file = ifcopenshell.file()
        self._schema = ifcopenshell.ifcopenshell_wrapper.schema_by_name(self._file.schema)
        self._entitymap = {}
        self._geometrymap = {}
        self._stylemap = {}

    def open(self, filepath: str):
        self.filepath = filepath
        self._file = ifcopenshell.open(filepath)
        self._schema = ifcopenshell.ifcopenshell_wrapper.schema_by_name(self._file.schema)
        self._entitymap = {}
        print("Opened file: {}".format(filepath))
        self.load_geometries()

    def get_entity(self, entity: ifcopenshell.entity_instance):
        """
        Returns the compas_ifc entity corresponding to the given ifcopenshell entity.

        Parameters
        ----------
        entity : ifcopenshell.entity_instance
            The ifcopenshell entity.

        Returns
        -------
        :class:`compas_ifc.entities.entity.Entity`
            The compas_ifc entity.
        """
        return self._entitymap.setdefault(entity.id(), Entity.factory(entity, self.model, self.entity_types))

    def get_entities_by_type(self, entity_type, accept_subtypes=True):
        """
        Returns all the entities of the given type.

        Parameters
        ----------
        entity_type : str
            The type of the entities to return.
        accept_subtypes : bool, optional
            Whether to include subtypes of the given type.
            Defaults to True.

        Returns
        -------
        List[:class:`compas_ifc.entities.entity.Entity`]
            The entities of the given type.
        """
        _entities = self._file.by_type(entity_type, include_subtypes=accept_subtypes)
        return [self.get_entity(_entity) for _entity in _entities]

    def get_entities_by_name(self, entity_name) -> List[Entity]:
        """Returns all the entities with the given name."""
        entities = self.get_entities_by_type("IfcRoot")
        return [entity for entity in entities if entity.name == entity_name]

    def get_entity_by_global_id(self, global_id) -> Entity:
        """Returns the entity with the given global id."""
        ifc_entity = self._file.by_guid(global_id)
        if ifc_entity:
            return self.get_entity(ifc_entity)

    def get_entity_by_id(self, id) -> Entity:
        """Returns the entity with the given id."""
        ifc_entity = self._file.by_id(id)
        if ifc_entity:
            return self.get_entity(ifc_entity)

    def get_all_entities(self) -> List[Entity]:
        """Returns all the entities in the model."""
        return [self.get_entity(_entity) for _entity in self._file]

    def file_size(self):
        """Returns the size of the IFC file in bytes."""
        file_stats = os.stat(self.filepath)
        size_in_mb = file_stats.st_size / (1024 * 1024)
        size_in_mb = round(size_in_mb, 2)
        return size_in_mb

    def get_preloaded_geometry(self, entity):
        return self._geometrymap.get(entity._entity.id())

    def get_preloaded_style(self, entity):
        return self._stylemap.get(entity._entity.id(), {})

    def load_geometries(self, include=None, exclude=None):
        """Load all the geometries of the IFC file using a fast multithreaded iterator."""
        print("Loading geometries...")
        import ifcopenshell.geom

        settings = ifcopenshell.geom.settings()
        if self.use_occ:
            settings.set(settings.USE_PYTHON_OPENCASCADE, True)

        iterator = ifcopenshell.geom.iterator(settings, self._file, multiprocessing.cpu_count(), include=include, exclude=exclude)
        start = time.time()
        if iterator.initialize():
            while True:
                shape = iterator.get()
                if self.use_occ:
                    from compas_occ.brep import OCCBrep

                    brep = OCCBrep.from_shape(shape.geometry)

                    shellcolors = []
                    for style_id, style in zip(shape.style_ids, shape.styles):
                        if style_id == -1:
                            shellcolors.append((0.5, 0.5, 0.5, 1.0))
                        else:
                            shellcolors.append(style)

                    self._geometrymap[shape.data.id] = brep
                    self._stylemap[shape.data.id] = {"shellcolors": shellcolors, "use_rgba": True}

                else:
                    from .brep import TessellatedBrep

                    matrix = shape.transformation.matrix.data
                    faces = shape.geometry.faces
                    edges = shape.geometry.edges
                    verts = shape.geometry.verts

                    matrix = np.array(matrix).reshape((4, 3))
                    matrix = np.hstack([matrix, np.array([[0], [0], [0], [1]])])
                    matrix = matrix.transpose()
                    transformation = Transformation.from_matrix(matrix.tolist())

                    facecolors = []
                    for m_id in shape.geometry.material_ids:
                        if m_id == -1:
                            facecolors.append([0.5, 0.5, 0.5, 1])
                            facecolors.append([0.5, 0.5, 0.5, 1])
                            facecolors.append([0.5, 0.5, 0.5, 1])
                            continue
                        material = shape.geometry.materials[m_id]
                        color = (*material.diffuse, 1 - material.transparency)
                        facecolors.append(color)
                        facecolors.append(color)
                        facecolors.append(color)

                    brep = TessellatedBrep(vertices=verts, edges=edges, faces=faces)

                    brep.transform(transformation)
                    self._geometrymap[shape.id] = brep
                    self._stylemap[shape.id] = {"facecolors": facecolors}

                if not iterator.next():
                    break

        print(f"Time to load all {len(self._geometrymap)} geometries {(time.time() - start):.3f}s")



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from compas_ifc.entities.generated.IFC4 import IfcBuilding
    from compas_ifc.entities.generated.IFC4 import IfcBuildingElement
    from compas_ifc.entities.generated.IFC4 import IfcBuildingStorey
    from compas_ifc.entities.generated.IFC4 import IfcGeographicElement
else:
    IfcBuilding = object


class IfcBuilding(IfcBuilding):
    @property
    def building_elements(self) -> list["IfcBuildingElement"]:
        return self.children_by_type("IfcBuildingElement", recursive=True)

    @property
    def geographic_elements(self) -> list["IfcGeographicElement"]:
        return self.children_by_type("IfcGeographicElement", recursive=True)

    @property
    def storeys(self) -> list["IfcBuildingStorey"]:
        return self.children_by_type("IfcBuildingStorey", recursive=True)



try:
    import numpy as np
    from compas.colors import Color
    from compas.tolerance import TOL
    from compas_occ.brep import OCCBrep
    from compas_viewer.scene.brepobject import BRepObject

    class IFCBrepObject(BRepObject):
        def __init__(self, shellcolors=None, **kwargs):
            super().__init__(**kwargs)
            self.shells = [shell.to_tesselation(TOL.lineardeflection)[0] for shell in self.brep.shells]
            self.shellcolors = shellcolors or [self.facecolor.rgba for _ in self.shells]
            self._bounding_box_center = None

        @property
        def brep(self) -> OCCBrep:
            return self.item

        @property
        def bounding_box_center(self):
            if not self._bounding_box_center:
                self._bounding_box_center = np.mean(self.points, axis=0)
            return self._bounding_box_center

        def _read_frontfaces_data(self):
            positions = []
            elements = np.array([], dtype=int).reshape(0, 3)
            colors = []

            for shell, color in zip(self.shells, self.shellcolors):
                shell_positions, shell_elements = shell.to_vertices_and_faces()
                shell_elements = np.array(shell_elements) + len(positions)
                positions += shell_positions
                elements = np.vstack((elements, shell_elements))
                colors += [Color(*color)] * len(shell_positions)  # TODO: this is terrible
                if color[3] < 1:
                    self.opacity = 0.999  # NOTE: this is to trigger the object order sorting

            return positions, colors, elements

        def _read_backfaces_data(self):
            positions = []
            elements = np.array([], dtype=int).reshape(0, 3)
            colors = []

            for shell, color in zip(self.shells, self.shellcolors):
                shell_positions, shell_elements = shell.to_vertices_and_faces()
                for element in shell_elements:
                    element.reverse()
                shell_elements = np.array(shell_elements) + len(positions)
                positions += shell_positions
                elements = np.vstack((elements, shell_elements))
                colors += [Color(*color)] * len(shell_positions)

            return positions, colors, elements

except ImportError:
    pass



import ifcopenshell

schema = ifcopenshell.ifcopenshell_wrapper.schema_by_name("IFC4")

init_template = "from .CLASS_NAME import CLASS_NAME\n"

template = """class CLASS_NAME(TYPE_NAME):
    \"\"\"Wrapper class for CLASS_NAME.\"\"\"
"""

TYPE_MAP = {
    "DOUBLE": "float",
    "INT": "int",
    "STRING": "str",
    "LOGICAL": "bool",
    "BOOL": "bool",
    "BINARY": "bytes",
}


if __name__ == "__main__":
    init_string = ""

    for declaration in schema.declarations():
        class_string = template

        if declaration.as_type_declaration():
            name = declaration.name()

            print(name)

            class_string = class_string.replace("CLASS_NAME", name)

            ifc_type = declaration.argument_types()[0]
            if ifc_type.startswith("AGGREGATE OF"):
                value_type = ifc_type.split("OF ")[1]
                if value_type == "ENTITY INSTANCE":
                    python_type = "list"  # TODO: handle this
                else:
                    python_type = f"list[{TYPE_MAP[value_type]}]"
            else:
                python_type = TYPE_MAP[ifc_type]

            class_string = class_string.replace("TYPE_NAME", python_type)

            print(class_string)
            # break

            with open(f"src/compas_ifc/entities/generated/{name}.py", "w") as f:
                f.write(class_string)

    # with open(f"src/compas_ifc/entities/generated/__init__.py", "w") as f:
    #     f.write(init_string)



import ifcopenshell

schema = ifcopenshell.ifcopenshell_wrapper.schema_by_name("IFC4")


template = """from enum import Enum

class CLASS_NAME(Enum):
    \"\"\"Wrapper class for CLASS_NAME.\"\"\"
"""

enum_template = "    NAME=INDEX\n"

if __name__ == "__main__":
    init_string = ""

    for declaration in schema.declarations():
        class_string = template

        if declaration.as_enumeration_type():
            name = declaration.name()
            class_string = class_string.replace("CLASS_NAME", name)

            for index, item in enumerate(declaration.enumeration_items()):
                enum_string = enum_template
                enum_string = enum_string.replace("NAME", item)
                enum_string = enum_string.replace("INDEX", str(index))
                class_string += enum_string

            print(class_string)
            # break

            with open(f"src/compas_ifc/entities/generated/{name}.py", "w") as f:
                f.write(class_string)



"""
********************************************************************************
entities
********************************************************************************

.. currentmodule:: compas_ifc.entities

Classes representing rooted entities (entities inheriting from IfcRoot).
Three types of entities are of particular interest:

* IfcContext :: **IfcProject**
* IfcObject :: IfcProduct :: IfcElement :: **IfcSpatialElement**

  * IfcSite
  * IfcBuilding
  * IfcBuildingStorey
  * IfcSpace

* IfcObject :: IfcProduct :: IfcElement :: **IfcBuildingElement**

  * IfcBeam
  * IfcColumn
  * IfcDoor
  * IfcRoof
  * IfcSlab
  * IfcStair
  * IfcWall
  * IfcWindow


Associations
============

Classifications
---------------

See also :ifc:`classification`.

Objects like building and spatial elements can be further described by associating references to external sources of information.
The source of information can be:

* a classification system;
* a dictionary server;
* any external catalogue that classifies the object further;
* a service that combine the above features.

These "associations" are called "classifications".

.. code-block:: python

    for association in entity.HasAssociations:
        if association.is_a("IfcRelAssociatesClassification"):
            ...

Material Associations
---------------------

See also :ifc:`material-association`.

Material associations indicate the physical composition of an object.
Materials can have representations for surface styles indicating colors, textures, and light reflectance for 3D rendering.
Materials can have representations for fill styles indicating colors, tiles, and hatch patterns for 2D rendering.
Materials can have properties such as density, elasticity, thermal resistance, and others as defined in this specification.
Materials can also be classified according to a referenced industry standard.

.. code-block:: python

    for association in entity.HasAssociations:
        if association.is_a("IfcRelAssociatesMaterial"):
            ...

Object Composition
==================

Objects may be composed into parts to indicate levels of detail, such as a building having multiple storeys, a framed wall having studs, or a task having subtasks.
Composition may form a hierarchy of multiple levels, where an object must have single parent,
or if a top-level object then declared within the single project or a project library.

Element Composition
-------------------

See also :ifc:`element-composition`.

Provision of an aggregation structure where the element is part of another element representing the composite.
The part then provides, if such concepts are in scope of the Model View Definition, exclusively the following:

* Body Geometry — The partial body shape representation and its placement;
* Material — The material information for the part.

The part may also provide, in addition to the aggregate, more specifically the following:

* Property Sets — The parts may have individual property sets assigned, solely or in addition to the composite;
* Quantity Sets — The parts may have individual quantity sets assigned, solely or in addition to the composite.

The part should not be contained in the spatial hierarchy, i.e. the concept Spatial Containment shall not be used at the level of parts.
The part is contained in the spatial structure by the spatial containment of its composite.

.. code-block:: python

    for element in entity.ContainsElements:
        # ...


Spatial Composition
-------------------

...

.. code-block:: python

    for element in entity.IsDecomposedBy:
        # ...


Classes
=======

Bases
-----

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Entity

Context
-------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Project

Spatial Elements
----------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Site
    Building
    BuildingStorey
    Space

Building Elements
-----------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Beam
    Column
    Door
    Roof
    Slab
    Stair
    Wall
    Window

"""

from .entity import Entity
from .root import Root
from .objectdefinition import ObjectDefinition
from .element import Element
from .spatialelement import SpatialElement
from .buildingelements import Beam
from .buildingelements import Column
from .buildingelements import Door
from .buildingelements import Roof
from .buildingelements import Railing
from .buildingelements import Slab
from .buildingelements import Stair
from .buildingelements import Wall
from .buildingelements import Window
from .buildingelements import BuildingElementProxy

from .space import Space
from .buildingstorey import BuildingStorey
from .building import Building
from .site import Site
from .geographicelement import GeographicElement

from .project import Project

DEFAULT_ENTITY_TYPES = {
    "IfcRoot": Root,
    "IfcObjectDefinition": ObjectDefinition,
    "IfcElement": Element,
    "IfcSpatialElement": SpatialElement,
    "IfcBeam": Beam,
    "IfcColumn": Column,
    "IfcDoor": Door,
    "IfcRoof": Roof,
    "IfcRailing": Railing,
    "IfcSlab": Slab,
    "IfcStair": Stair,
    "IfcWall": Wall,
    "IfcWindow": Window,
    "IfcSpace": Space,
    "IfcBuildingStorey": BuildingStorey,
    "IfcBuilding": Building,
    "IfcSite": Site,
    "IfcProject": Project,
    "IfcBuildingElementProxy": BuildingElementProxy,
    "IfcGeographicElement": GeographicElement,
}


def factory(entity, model, entity_types=None) -> Entity:
    """Factory function for creating an compas_ifc entity object from an Ifc entity, the function finds closest matched class from the bottom of inherentance.

    Parameters
    ----------
    entity : IfcEntity
        An IfcEntity object.
    model : IfcModel
        An IfcModel object.

    Returns
    -------
    Entity
        An Entity object.

    """
    entity_types = entity_types or DEFAULT_ENTITY_TYPES

    declaration = model.schema.declaration_by_name(entity.is_a())
    inheritance = [declaration.name()]
    while hasattr(declaration, "supertype") and declaration.supertype():
        inheritance.append(declaration.supertype().name())
        declaration = declaration.supertype()

    for ifc_type in inheritance:
        if ifc_type in entity_types:
            return entity_types[ifc_type](entity, model)

    dynamic_class = type(entity.is_a(), (Entity,), {})
    return dynamic_class(entity, model)


Entity.factory = staticmethod(factory)



# from compas_occ.geometry import OCCNurbsCurve
from compas.geometry import Frame
from compas.geometry import Line
from compas.geometry import Plane
from compas.geometry import Point
from compas.geometry import Vector

# from compas.geometry import Transformation


def IfcCartesianPoint_to_point(cartesian_point) -> Point:
    """
    Convert an IFC CartesianPoint [cartesianpoint]_ to a COMPAS point.

    """
    return Point(*cartesian_point.Coordinates)


def IfcDirection_to_vector(direction) -> Vector:
    """
    Convert an IFC Direction [direction]_ to a COMPAS vector.

    """
    return Vector(*direction.DirectionRatios)


def IfcVector_to_vector(vector) -> Vector:
    """
    Convert an IFC Vector [vector]_ to a COMPAS vector.

    """
    direction = IfcDirection_to_vector(vector.Orientation)
    direction.scale(vector.Magnitude)
    return direction


def IfcLine_to_line(line) -> Line:
    """
    Convert an IFC Line [line]_ to a COMPAS line.

    """
    point = IfcCartesianPoint_to_point(line.Pnt)
    vector = IfcDirection_to_vector(line.Dir)
    return Line(point, point + vector)


def IfcPlane_to_plane(plane) -> Plane:
    """
    Convert an IFC Plane [plane]_ to a COMPAS plane.

    """
    point = IfcCartesianPoint_to_point(plane.Position.Location)
    normal = IfcDirection_to_vector(plane.Position.P[3])
    return Plane(point, normal)


def IfcAxis2Placement2D_to_frame(placement, context=None) -> Frame:
    """
    Convert an IFC Axis2Placement2D [axis2placement2d]_ to a COMPAS frame.

    An Axis2Placement2D is a 2D placement based on a frame defined by a point and 2 vectors.

    """
    # use the coordinate system of the representation context to replace missing axes
    # use defaults if also those not available
    point = IfcCartesianPoint_to_point(placement.Location)
    zaxis = Vector.Zaxis()
    if placement.RefDirection:
        xaxis = IfcDirection_to_vector(placement.RefDirection)
    else:
        xaxis = Vector.Xaxis()
    yaxis = zaxis.cross(xaxis)
    return Frame(point, xaxis, yaxis)


def IfcAxis2Placement3D_to_frame(placement, context=None) -> Frame:
    """
    Convert an IFC Axis2Placement3D [axis2placement3d]_ to a COMPAS frame.

    An Axis2Placement3D is a 3D placement based on a frame defined by a point and 2 vectors.

    """
    # use the coordinate system of the representation context to replace missing axes
    # use defaults if also those not available
    point = IfcCartesianPoint_to_point(placement.Location)
    if placement.Axis:
        zaxis = IfcDirection_to_vector(placement.Axis)
    else:
        zaxis = Vector.Zaxis()
    if placement.RefDirection:
        xaxis = IfcDirection_to_vector(placement.RefDirection)
    else:
        xaxis = Vector.Xaxis()
    yaxis = zaxis.cross(xaxis)
    xaxis = yaxis.cross(zaxis)
    return Frame(point, xaxis, yaxis)


# this needs to output a transformation
# not a frame
def IfcCartesianTransformationOperator3D_to_frame(operator) -> Frame:
    """
    Convert an IFC CartesianTransformationOperator3D to a COMPAS Frame.

    """
    Axis1 = operator.Axis1
    Axis2 = operator.Axis2
    # Axis3 = operator.Axis3
    LocalOrigin = operator.LocalOrigin
    # Scale = operator.Scale

    xaxis = Vector.Xaxis() if not Axis1 else IfcDirection_to_vector(Axis1)
    yaxis = Vector.Yaxis() if not Axis2 else IfcDirection_to_vector(Axis2)
    # zaxis = Vector.Zaxis() if not Axis3 else IfcDirection_to_vector(Axis3)
    point = IfcCartesianPoint_to_point(LocalOrigin)
    return Frame(point, xaxis, yaxis)


def IfcProfileDef_to_curve(profile_def):
    from compas_occ.brep import OCCBrepFace

    pd = profile_def

    if pd.is_a("IfcParameterizedProfileDef"):
        if pd.is_a("IfcRectangleProfileDef"):
            frame = IfcAxis2Placement2D_to_frame(pd.Position)
            points = [
                frame.point + frame.xaxis * +0.5 * pd.XDim + frame.yaxis * -0.5 * pd.YDim,
                frame.point + frame.xaxis * +0.5 * pd.XDim + frame.yaxis * +0.5 * pd.YDim,
                frame.point + frame.xaxis * -0.5 * pd.XDim + frame.yaxis * +0.5 * pd.YDim,
                frame.point + frame.xaxis * -0.5 * pd.XDim + frame.yaxis * -0.5 * pd.YDim,
            ]
            return OCCBrepFace.from_polygon(points)

        else:
            raise NotImplementedError(pd)
    elif pd.is_a("IfcArbitraryClosedProfileDef"):
        return IfcCurve_to_face(pd.OuterCurve)
    elif pd.is_a("IfcCompositeProfileDef"):
        faces = []
        for profile in pd.Profiles:
            if profile.is_a("IfcArbitraryClosedProfileDef"):
                faces.append(IfcCurve_to_face(profile.OuterCurve))
            else:
                raise NotImplementedError(profile)
        return faces
    else:
        raise NotImplementedError(pd)


def IfcCurve_to_face(curve):
    from compas_occ.brep import OCCBrepFace

    if curve.is_a("IfcIndexedPolyCurve"):
        points = [(x, y, 0) for x, y in curve.Points[0]]
        return OCCBrepFace.from_polygon(points)
    elif curve.is_a("IfcPolyline"):
        points = [IfcCartesianPoint_to_point(pt) for pt in curve.Points]
        return OCCBrepFace.from_polygon(points)
    else:
        raise NotImplementedError(curve)



from typing import TYPE_CHECKING

from compas.geometry import Scale

from compas_ifc.resources import IfcLocalPlacement_to_transformation  # TODO: move this
from compas_ifc.resources import IfcShape_to_brep  # TODO: move this
from compas_ifc.resources import IfcShape_to_tessellatedbrep  # TODO: move this

if TYPE_CHECKING:
    from compas_ifc.entities.generated.IFC4 import IfcElement


def entity_transformation(entity: "IfcElement"):
    """
    Construct the transformation of an entity.
    """
    scale = entity.model.project.length_scale

    if entity.ObjectPlacement:
        scaled_placement = IfcLocalPlacement_to_transformation(entity.ObjectPlacement, scale=scale)
    else:
        scaled_placement = Scale.from_factors([scale, scale, scale])

    return scaled_placement


def entity_body_geometry(entity: "IfcElement", context="Model", use_occ=False, apply_transformation=True):
    """
    Construct the body geometry representations of an entity.
    """
    bodies = []

    representation = None
    if entity.Representation:
        for temp in entity.Representation.Representations:
            if temp.RepresentationIdentifier == "Body" and temp.ContextOfItems.ContextType == context:
                representation = temp
                break

    if not representation:
        return bodies

    for item in representation.Items:
        if use_occ:
            brep = IfcShape_to_brep(item.entity)  # TODO: improve this, should just give item
            bodies.append(brep)
        else:
            tessellatedbrep = IfcShape_to_tessellatedbrep(item.entity)  # TODO: improve this, should just give item
            bodies.append(tessellatedbrep)

    if apply_transformation:
        transformation = entity_transformation(entity)
        for body in bodies:
            body.transform(transformation)

    return bodies


def entity_opening_geometry(entity: "IfcElement", use_occ=False, apply_transformation=True):
    """
    Construct the opening geometry representations of an entity.
    """

    voids = []
    for opening in entity.HasOpenings():
        element = opening.RelatedOpeningElement

        if apply_transformation:
            transformation = entity_transformation(element)

        for representation in element.Representation.Representations:
            for item in representation.Items:
                if use_occ:
                    brep = IfcShape_to_brep(item.entity)  # TODO: improve this, should just give item
                    if apply_transformation:
                        brep.transform(transformation)
                    voids.append(brep)
                else:
                    tessellatedbrep = IfcShape_to_tessellatedbrep(item.entity)  # TODO: improve this, should just give item
                    if apply_transformation:
                        tessellatedbrep.transform(transformation)
                    voids.append(tessellatedbrep)

    return voids



from typing import TYPE_CHECKING

from compas_ifc.resources import IfcLocalPlacement_to_frame
from compas_ifc.resources.representation import write_body_representation
from compas_ifc.resources.shapes import frame_to_ifc_axis2_placement_3d

if TYPE_CHECKING:
    from compas_ifc.entities.generated.IFC4 import IfcProduct
else:
    IfcProduct = object


class IfcProduct(IfcProduct):
    @property
    def style(self):
        return self.file.get_preloaded_style(self)

    @property
    def geometry(self):
        if not getattr(self, "_geometry", None):
            self._geometry = self.file.get_preloaded_geometry(self)
            if self.frame and self._geometry:
                # NOTE: preloaded geometry is pre-transformed because of boolean.
                # The pre-transformation is not necessarily the same as the frame of entity.
                # Therefore, we need to re-transform the geometry back to its original location.
                T = self.frame.to_transformation()
                self._geometry.transform(T.inverse())
        return self._geometry

    @geometry.setter
    def geometry(self, geometry):
        self._geometry = geometry
        # Update the representation in the IFC file
        # TODO: delete existing representation
        # TODO: make this function more transparent
        write_body_representation(self.file._file, geometry, self.entity, self.file.default_body_context.entity)

    @property
    def frame(self):
        if not getattr(self, "_frame", None):
            if self.ObjectPlacement:
                self._frame = IfcLocalPlacement_to_frame(self.ObjectPlacement)
            else:
                self._frame = None

        # print(self._frame)
        return self._frame

    @frame.setter
    def frame(self, frame):
        self._frame = frame
        # Update the placement in the IFC file
        # TODO: consider parent frame
        # TODO: make this function more transparent
        loacal_placement = frame_to_ifc_axis2_placement_3d(self.file._file, frame)
        placement = self.file._create_entity("IfcLocalPlacement", RelativePlacement=loacal_placement)
        self.ObjectPlacement = placement



from compas.datastructures import Tree
from compas.datastructures import TreeNode

from .root import Root


class ObjectDefinition(Root):
    """Base class for all object definitions. An object definition is a definition of a thing that is or may be part of a spatial structure.

    Attributes
    ----------
    parent : :class:`compas_ifc.entities.ObjectDefinition`
        The parent of this element in spatial hierarchy.
    children : List[:class:`compas_ifc.entities.ObjectDefinition`]
        The children of this element in spatial hierarchy.

    """

    def __init__(self, entity, model) -> None:
        super().__init__(entity, model)
        self._parent = None

    def decomposes(self):
        """Return the relation that decomposes this element."""
        if self not in self.model._new_entities:
            for rel in self._entity.Decomposes:
                return self.model.reader.get_entity(rel)

    @property
    def parent(self):
        if not self._parent and self._entity:
            relation = self.decomposes()
            if relation:
                self._parent = relation["RelatingObject"]
        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent

    def is_decomposed_by(self):
        """Return the relation that this element is decomposed by."""
        return [self.model.reader.get_entity(rel) for rel in self._entity.IsDecomposedBy]

    @property
    def children(self):
        children = [entity for entity in self.model.get_entities_by_type("IfcObjectDefinition") if entity.parent == self]
        for entity in self.model._new_entities:
            if entity.parent == self and entity not in children:
                children.append(entity)

        # sort children by name
        children.sort(key=lambda x: x.name)
        return children

    def traverse(self, recursive: bool = True):
        """Traverse children of this element.

        Parameters
        ----------
        recursive : bool, optional
            Whether to traverse down the tree recursively, by default True

        Yields
        ------
        :class:`compas_ifc.entities.ObjectDefinition`
            The children of this element.
        """
        for child in self.children:
            yield child
            if recursive:
                yield from child.traverse(recursive)

    def traverse_ancestor(self, recursive: bool = True):
        """Traverse ancestors of this element.

        Parameters
        ----------
        recursive : bool, optional
            Whether to traverse up the tree recursively, by default True

        Yields
        ------
        :class:`compas_ifc.entities.ObjectDefinition`
            The ancestors of this element.
        """
        parent = self.parent
        if parent:
            yield from parent.traverse_ancestor(recursive)
            yield parent

    def traverse_branch(self, recursive: bool = True):
        """Traverse the spatial branch of this element.

        Parameters
        ----------
        recursive : bool, optional
            Whether to traverse up and down the tree recursively, by default True

        Yields
        ------
        :class:`compas_ifc.entities.ObjectDefinition`
            The ancestors, self, and children of this element.
        """
        yield from self.traverse_ancestor(recursive)
        yield self
        yield from self.traverse(recursive)

    def print_spatial_hierarchy(self, max_level: int = 4) -> None:
        """Print the spatial hierarchy of this element.

        Parameters
        ----------
        max_level : int, optional
            The maximum level of the hierarchy to print, by default 4

        Returns
        -------
        None
        """
        print("=" * 80)
        print("SPATIAL HIERARCHY")
        print(self.spatial_hierarchy.get_hierarchy_string(max_depth=max_level))
        print("=" * 80)

    @property
    def spatial_hierarchy(self) -> Tree:
        tree = Tree()

        class Node(TreeNode):
            def __repr__(self):
                return str(self.name)

        root = Node(self)

        tree.add(root)

        def traverse(entity, parent):
            node = Node(entity)
            parent.add(node)
            for child in entity.children:
                traverse(child, node)

        traverse(self, root)
        return tree



from compas_ifc.entities.element import Element
from compas_ifc.entities.spatialelement import SpatialElement


class BuildingElement(Element):
    """
    Base class for all building elements.

    References
    ----------
    * Spatial Containment: :ifc:`spatial-containment`

    """

    def __init__(self, entity, model) -> None:
        super().__init__(entity, model)
        self._composite_body = None
        self._composite_opening = None
        self._composite_body_with_opening = None

    def is_contained(self) -> bool:
        """
        Verify that this building element is contained in a spatial structure.

        Returns
        -------
        bool

        """
        return len(self._entity.ContainedInStructure) == 1

    def is_containment_hierarchical(self) -> bool:
        """
        Verify that the spatial containment of this building element is hierarchical,
        meaning that the element is contained in no more than one (1) spatial element.

        Returns
        -------
        bool

        """
        if len(self._entity.ContainedInStructure) > 1:
            return False
        return True

    def container(self) -> SpatialElement:
        """
        Get the container of the building element.

        Containment is hierarchical, meaning that the element can only be contained in one spatial element.

        Returns
        -------
        :class:`SpatialElement`

        """
        return self.parent

    @property
    def building_element_parts(self):
        return [part for part in self.children if part.is_a("IfcBuildingElementPart")]

    @property
    def composite_body(self):
        bodies = []
        for part in self.building_element_parts:
            bodies.extend(part.body)
        self._composite_body = bodies
        return self._composite_body

    @composite_body.setter
    def composite_body(self, value):
        self._composite_body = value

    @property
    def composite_opening(self):
        voids = self.opening
        for part in self.building_element_parts:
            voids.extend(part.opening)
        self._composite_opening = voids
        return self._composite_opening

    @property
    def composite_body_with_opening(self):
        from compas_ifc.representation import entity_body_with_opening_geometry

        if not self._composite_body_with_opening:
            self._composite_body_with_opening = entity_body_with_opening_geometry(entity=self, bodies=self.composite_body, voids=self.composite_opening)
        return self._composite_body_with_opening

    @composite_body_with_opening.setter
    def composite_body_with_opening(self, value):
        self._composite_body_with_opening = value


class Beam(BuildingElement):
    pass


class Column(BuildingElement):
    pass


class Wall(BuildingElement):
    @property
    def gross_area(self):
        self.properties.get("GrossSideArea")


class Window(BuildingElement):
    pass


class Roof(BuildingElement):
    pass


class Slab(BuildingElement):
    pass


class Door(BuildingElement):
    pass


class Railing(BuildingElement):
    pass


class Stair(BuildingElement):
    pass


class BuildingElementProxy(BuildingElement):
    pass



import time

import ifcopenshell
from ifcopenshell.api import run

from .entities.element import Element
from .entities.entity import Entity
from .entities.objectdefinition import ObjectDefinition
from .entities.product import Product
from .entities.project import Project
from .entities.root import Root
from .resources.representation import write_body_representation
from .resources.shapes import frame_to_ifc_axis2_placement_3d


class IFCWriter(object):
    """
    A class for writing IFC files.

    Parameters
    ----------
    model : :class:`compas_ifc.model.Model`
        The model to which the writer belongs.

    Attributes
    ----------
    model : :class:`compas_ifc.model.Model`
        The model to which the writer belongs.
    file : :class:`ifcopenshell.file`
        The IFC file to which the model is being written to.
    default_context : :class:`ifcopenshell.entity_instance`
        The default context of the model. Created if it does not exist.
    default_body_context : :class:`ifcopenshell.entity_instance`
        The default body context of the model. Created if it does not exist.
    default_project : :class:`ifcopenshell.entity_instance`
        The default project of the model. Created if it does not exist.
    default_site : :class:`ifcopenshell.entity_instance`
        The default site of the model. Created if it does not exist.
    default_building : :class:`ifcopenshell.entity_instance`
        The default building of the model. Created if it does not exist.
    default_building_storey : :class:`ifcopenshell.entity_instance`
        The default building storey of the model. Created if it does not exist.

    """

    def __init__(self, model):
        self.file = None
        self.model = model
        self._entitymap = {}
        self._relationmap_aggregates = {}
        self._relationmap_contains = {}
        self._representationmap = {}
        self._psetsmap = {}
        self._default_context = None
        self._default_body_context = None
        self._default_project = None
        self._default_site = None
        self._default_building = None
        self._default_building_storey = None
        self._default_owner_history = None

    @property
    def default_context(self):
        # TODO: allow loading of existing contexts
        if not self._default_context:
            self._default_context = run("context.add_context", self.file, context_type="Model")
        return self._default_context

    @property
    def default_body_context(self):
        # TODO: allow loading of existing contexts
        if not self._default_body_context:
            self._default_body_context = run(
                "context.add_context",
                self.file,
                context_type="Model",
                context_identifier="Body",
                target_view="MODEL_VIEW",
                parent=self.default_context,
            )
        return self._default_body_context

    @property
    def default_project(self):
        if not self._default_project:
            if not self.model.projects:
                self._default_project = self.file.create_entity("IfcProject", GlobalId=self.create_guid(), Name="Default Project")
                run("unit.assign_unit", self.file)
            else:
                self._default_project = self.write_entity(self.model.projects[0])
        return self._default_project

    @property
    def default_site(self):
        if not self._default_site:
            if not self.model.sites:
                self._default_site = self.file.create_entity("IfcSite", Name="Default Site", GlobalId=self.create_guid())
                self.file.create_entity(
                    "IfcRelAggregates",
                    GlobalId=self.create_guid(),
                    RelatingObject=self.default_project,
                    RelatedObjects=[self._default_site],
                )
            else:
                self._default_site = self.write_entity(self.model.sites[0])
        return self._default_site

    @property
    def default_building(self):
        if not self._default_building:
            if not self.model.buildings:
                self._default_building = self.file.create_entity("IfcBuilding", GlobalId=self.create_guid(), Name="Default Building")
                self.file.create_entity(
                    "IfcRelAggregates",
                    GlobalId=self.create_guid(),
                    RelatingObject=self.default_site,
                    RelatedObjects=[self._default_building],
                )
            else:
                self._default_building = self.write_entity(self.model.buildings[0])
        return self._default_building

    @property
    def default_building_storey(self):
        if not self._default_building_storey:
            if not self.model.building_storeys:
                self._default_building_storey = self.file.create_entity("IfcBuildingStorey", GlobalId=self.create_guid(), Name="Default Storey")
                self.file.create_entity(
                    "IfcRelAggregates",
                    GlobalId=self.create_guid(),
                    RelatingObject=self.default_building,
                    RelatedObjects=[self._default_building_storey],
                )
            else:
                self._default_building_storey = self.write_entity(self.model.building_storeys[0])
        return self._default_building_storey

    @property
    def default_owner_history(self):
        if not self._default_owner_history:
            import compas_ifc

            person = self.file.create_entity("IfcPerson")
            organization = self.file.create_entity("IfcOrganization", Name="compas.dev")
            person_and_org = self.file.create_entity("IfcPersonAndOrganization", ThePerson=person, TheOrganization=organization)
            application = self.file.create_entity(
                "IfcApplication",
                ApplicationDeveloper=organization,
                Version=compas_ifc.__version__,
                ApplicationFullName="compas_ifc",
                ApplicationIdentifier="compas_ifc v" + compas_ifc.__version__,
            )

            owner_history = self.file.create_entity(
                "IfcOwnerHistory",
                OwningUser=person_and_org,
                OwningApplication=application,
                ChangeAction="ADDED",
                CreationDate=int(time.time()),
            )

            self._default_owner_history = owner_history
        return self._default_owner_history

    def create_guid(self):
        return ifcopenshell.guid.new()

    def reset(self):
        """Resets the writer to start with a new ifc file."""
        self.file = ifcopenshell.file(schema=self.model.schema.name())
        self._entitymap = {}
        self._default_context = None
        self._default_body_context = None
        self._default_project = None
        self._default_site = None
        self._default_building = None
        self._default_building_storey = None

    def export(self, entities, filepath: str) -> None:
        """Exports the given entities to the ifc file."""
        print("Exporting IFC file to: " + filepath)
        self.reset()
        # TODO: make this better
        self.default_project

        # TODO: make this better
        for entity in self.model._new_entities:
            if not entity.parent and self.model.building_storeys:
                entity.parent = self.model.building_storeys[0]

        for entity in entities:
            for node in entity.traverse_branch():
                self.write_entity(node)

        for entity in entities:
            for node in entity.traverse_branch():
                self.write_relation(node)

        self.file.write(filepath)
        print("Done.")

    def save(self, filepath: str) -> None:
        """Writes the model as ifc file to the given filepath."""
        print("Saving IFC file to: " + filepath)
        self.reset()
        # TODO: make this better
        self.default_project

        for entity in self.model.get_all_entities():
            self.write_entity(entity)

        for entity in self.model._new_entities:  # TODO: needs better api
            self.write_relation(entity)

        self.file.write(filepath)
        print("Done.")

    def write_relation(self, entity: Entity):
        """Writes the relation of the given entity to the ifc file."""
        if entity._entity:
            # Entity from ifc file
            self.trim_existing_relation(entity)
        else:
            # Entity created in memory
            if isinstance(entity, ObjectDefinition) and not isinstance(entity, Project):
                # Only ObjectDefinitions can have parental relations, except for Project.
                self.create_new_relation(entity)

    def trim_existing_relation(self, entity: Entity) -> None:
        """writes a existing parental relation to the ifc file, trimming the non existing children."""

        if not entity.parent:
            return

        chilren_to_include = []
        relation = None

        # Modify the relation to only include the children that has been written to the ifc file.
        if hasattr(entity, "contained_in_structure"):
            relation = entity.contained_in_structure()
            for child in relation["RelatedElements"]:
                if child in self._entitymap:
                    chilren_to_include.append(child)
            relation["RelatedElements"] = chilren_to_include
        elif hasattr(entity, "decomposes"):
            relation = entity.decomposes()
            for child in relation["RelatedObjects"]:
                if child in self._entitymap:
                    chilren_to_include.append(child)
            relation["RelatedObjects"] = chilren_to_include
        else:
            raise Exception("Entity has no parent relation.")

        self.write_entity(relation)

    def create_new_relation(self, entity: ObjectDefinition):
        """Writes a new parental relation to the ifc file."""

        if entity.parent:
            parent = self._entitymap[entity.parent]
        else:
            # If no parent is given, use the default building storey.
            parent = self.default_building_storey

        child = self._entitymap[entity]

        if isinstance(entity, Element):
            if self._relationmap_contains.get(parent):
                relation = self._relationmap_contains[parent]
                children = set(relation.RelatedElements)
                children.add(child)
                children = list(children)
                children.sort(key=lambda x: x.Name)
                relation.RelatedElements = tuple(children)
            else:
                relation = self.file.create_entity(
                    "IfcRelContainedInSpatialStructure",
                    GlobalId=self.create_guid(),
                    RelatingStructure=parent,
                    RelatedElements=[child],
                )
                self._relationmap_contains[parent] = relation
        else:
            if self._relationmap_aggregates.get(parent):
                relation = self._relationmap_aggregates[parent]
                children = set(relation.RelatedObjects)
                children.add(child)
                children = list(children)
                children.sort(key=lambda x: x.Name)
                relation.RelatedObjects = tuple(children)
            else:
                relation = self.file.create_entity("IfcRelAggregates", GlobalId=self.create_guid(), RelatingObject=parent, RelatedObjects=[child])
                self._relationmap_aggregates[parent] = relation

    def write_entity(self, entity: Entity) -> None:
        """Writes the given entity recursively with all its referencing attributes to the ifc file."""
        if entity in self._entitymap:
            return self._entitymap[entity]

        attributes = {}
        if isinstance(entity, Root):
            attributes["OwnerHistory"] = self.default_owner_history

        for key, value in entity.attributes.items():
            if isinstance(value, Entity):
                attributes[key] = self.write_entity(value)
            elif isinstance(value, list) and value and isinstance(value[0], Entity):
                attributes[key] = [self.write_entity(v) for v in value]
            else:
                attributes[key] = value

        # ifc_entity = self.file.create_entity(entity.ifc_type, **attributes)
        ifc_entity = self.file.create_entity(entity.ifc_type)
        for key, value in attributes.items():
            if value is not None:
                setattr(ifc_entity, key, value)

        self._entitymap[entity] = ifc_entity

        if not entity._entity:
            # Entity created in memory
            self.write_entity_representation(entity)
            self.write_entity_placement(entity)
            self.write_entity_pset(entity, ifc_entity)

        if isinstance(entity, Project):
            print("Writing project: " + str(entity))
            run("unit.assign_unit", self.file)

        return ifc_entity

    def write_entity_pset(self, entity: Entity, ifc_entity: ifcopenshell.entity_instance):
        for name, properties in entity.psets.items():
            if id(properties) in self._psetsmap:
                pset = self._psetsmap[id(properties)]
                # TODO: This can be mreged too.
                self.file.create_entity("IfcRelDefinesByProperties", GlobalId=self.create_guid(), RelatingPropertyDefinition=pset, RelatedObjects=[ifc_entity])
            else:
                pset = run("pset.add_pset", self.file, product=ifc_entity, name=name)
                self._psetsmap[id(properties)] = pset
                run("pset.edit_pset", self.file, pset=pset, properties=properties)

    def write_entity_representation(self, entity: Entity):
        """Writes the representations of the given entity to the ifc file."""
        if isinstance(entity, Product):
            if entity.body:
                if id(entity.body) not in self._representationmap:
                    representation = write_body_representation(self.file, entity.body, self._entitymap[entity], self.default_body_context)
                    self._representationmap[id(entity.body)] = representation
                else:
                    representation = self._representationmap[id(entity.body)]
                    run(
                        "geometry.assign_representation",
                        self.file,
                        product=self._entitymap[entity],
                        representation=representation,
                    )

    def write_entity_placement(self, entity: Entity):
        """Writes the placement of the given entity to the ifc file."""
        if isinstance(entity, Product):
            if entity.frame:
                # TODO: consider parent frame
                loacal_placement = frame_to_ifc_axis2_placement_3d(self.file, entity.frame)
                placement = self.file.create_entity("IfcLocalPlacement", RelativePlacement=loacal_placement)
                self._entitymap[entity].ObjectPlacement = placement



from typing import List

from compas_ifc.entities.building import Building
from compas_ifc.entities.geographicelement import GeographicElement
from compas_ifc.entities.spatialelement import SpatialElement


class Site(SpatialElement):
    """
    Class representing an IFC site.

    Attributes
    ----------
    buildings : List[:class:`Building`]
        The buildings contained in the site.
        If the buildings are not properly nested, they can be found at the top level of an IFC model.
    address : str
        The compiled address of the site.
    geographic_elements : List[:class:`GeographicElement`]
        The geographic elements contained in the site.

    """

    def __init__(self, entity, model):
        super().__init__(entity, model)
        self._buildings = None
        self._geographic_elements = None

    @property
    def buildings(self) -> List[Building]:
        return [entity for entity in self.traverse() if entity.is_a("IfcBuilding")]

    @property
    def geographic_elements(self) -> List[GeographicElement]:
        return [entity for entity in self.traverse() if entity.is_a("IfcGeographicElement")]

    @property
    def address(self) -> str:
        attr = self.attribute("SiteAddress")
        address = ""
        address += ", ".join(attr["AddressLines"])
        address += f", {attr['Country']}-{attr['PostalCode']}"
        address += f" {attr['Town']}"
        return address

    def add_building(self, building: Building) -> None:
        """
        Add a building to this site.

        Parameters
        ----------
        building : :class:`Building`

        Returns
        -------
        None

        """
        raise NotImplementedError
